import numpy as np
from arnie.bpps import bpps

def threshknot_util(sequence, package='vienna_2', theta=0):
    '''
    Inputs:
    sequence: RNA sequence
    package: folding package to use
    
    Set theta = 0 to not filter base pairs as in ThreshKnot.
    
    Returns: N x N matrix of base pair probabilities. Nonzero entries represent base pairs
    predicted in final (possibly pseudoknotted) structure.
    Probabilities are their associated probability (obvs).
    '''
    
    bp_matrix = bpps(sequence, package=package)
    
    # if desired, filter base pair probabilities below a cutoff
    bp_matrix[np.where(bp_matrix <= theta)] = 0
    output = np.zeros([len(sequence),len(sequence)])
    
    # ProbKnot heuristic part 1: get all base pairs where p(ij) == p_max(i)
    output[np.where(bp_matrix == np.max(bp_matrix,axis=0))] = 1
    
    # ProbKnot heuristic part 2: get all base pairs where p(ij) == p_max(j)
    array_of_bps = np.clip(output+np.transpose(output)-1,0,1)
    
    # setting all bp probabilities not corresponding to a final selected base pair to zero
    bp_matrix[np.where(array_of_bps == 0)] = 0
    
    return bp_matrix



from arnie.utils import *

samiv_struct = "((((....(.((((((....((.[[[[[)).)))))))(((..((((((..{{{{)).)))).)))]]]]]....))))..((((.(((((.......))))).))))....}}}}..."

hotknots = "((((((....(((((((......[[[[[..))))))).....(((((....{{{{{..)))))...]]]]]..))))))..((((.(((((((...))))))).))))...}}}}}..."
ipknot = "[[[[[[....[[[..........(((((((((((]]].((((.(((.((......)).))).)))).......]]]]]].......(((((((...)))))))..))))))).)))).."
knotty = "(((..[[[[[))).]].]]]...[[[[[[[[[[[[...[[[[.[[[.[[......]].]]].]]]]((.(((....))).))....(((((((...))))))).]]]]]]]].]]]].."
spotrna = ".(((((((....(((((...((.....[[.)))))))]](((.((((((((....)))).)))))))......))))))).......((((((((.))))))))(.........)...."
e2efold = "......(............((...(.......(.............)..........)).)........(.(......(.(.(.([...).]).)...)....).).....)......."
pknots = ".(((.......))).((.....))(((((((((((((((((((.((((((......)).))))))))...(((....))).)))...((((((.....)))))).)))))))).))))."
empty = "." * len(samiv_struct)


def test_is_pk():
    assert(is_PK(samiv_struct))
    assert(not is_PK(pknots))


def test_compare_struct():
    assert(compare_structure_to_native(hotknots, samiv_struct, metric="PPV") == 0.8205128205128205)
    assert(compare_structure_to_native(hotknots, samiv_struct, metric="sensitivity") == 0.8)
    assert(compare_structure_to_native(hotknots, samiv_struct, metric="F1_score") == 0.810126582278481)
    assert(compare_structure_to_native(hotknots, samiv_struct, metric="all")["F1_score"] == 0.810126582278481)
    assert(compare_structure_to_native(empty, samiv_struct, metric="all")["F1_score"] == 0)
    assert(compare_structure_to_native(hotknots, samiv_struct, metric="all", PK_involved=True)["F1_score"] == 0.7796610169491526)
    assert(compare_structure_to_native(hotknots, samiv_struct, metric="all", PK_involved=False)["F1_score"] == 0.9)


def test_compare_structs():
    assert(0.4266666666666667 == compare_structures_to_natives([hotknots, spotrna], [samiv_struct, samiv_struct], comparison="basepairs")['PPV'])
    assert(1.0 == compare_structures_to_natives([hotknots, spotrna], [samiv_struct, samiv_struct], comparison="is_PK")["F1_score"])
    assert(1.0 == compare_structures_to_natives([hotknots, spotrna, pknots, empty], [samiv_struct, samiv_struct, samiv_struct, samiv_struct], comparison="is_PK", metric="PPV"))
    assert(0.5 == compare_structures_to_natives([hotknots, spotrna, pknots, empty], [samiv_struct, samiv_struct, samiv_struct, samiv_struct], comparison="is_PK", metric="sensitivity"))
    assert(0.6666666666666666 == compare_structures_to_natives([hotknots, spotrna, pknots, empty], [samiv_struct, samiv_struct, samiv_struct, samiv_struct], comparison="is_PK", metric="F1_score"))
    assert(0.25 == compare_structures_to_natives([hotknots, spotrna, pknots, empty], [samiv_struct, samiv_struct, samiv_struct, samiv_struct], comparison="non_PK_basepairs")["sensitivity"])
    assert(compare_structures_to_natives([hotknots, spotrna, pknots, empty], [samiv_struct, samiv_struct, samiv_struct, samiv_struct], comparison="PK_basepairs")["F1_score"] == 0.28571428571428575)


if __name__ == '__main__':
    test_is_pk()
    test_compare_struct()
    test_compare_structs()



import os, re, sys
import subprocess as sp
import random, string
import numpy as np
from .utils import *
from .pfunc import pfunc

# load package locations from yaml file, watch! global dict
package_locs = load_package_locations()

def bpps(sequence, package='vienna', constraint=None, pseudo=False,
         T=37, coaxial=True, linear=False, dna=False,
        motif=None, dangles=True,param_file=None,reweight=None, beam_size=100, DEBUG=False, threshknot=False,
        probing_signal=None, probing_kws=None,DIRLOC=None):

    ''' Compute base pairing probability matrix for RNA sequence.

    Args:
    sequence (str): nucleic acid sequence
    T (float): temperature (Celsius)
    linear (bool): call LinearPartition to estimate Z in Vienna or Contrafold
    constraint (str): structure constraint (functional in vienna, contrafold, rnastructure)
    motif (str): argument to vienna motif
    pseudo (bool): (NUPACK only) include pseudoknot calculation
    dangles (bool): dangles or not, specifiable for vienna, nupack
    dna (bool): (NUPACK only) use SantaLucia 1998 parameters for DNA
    coaxial (bool): coaxial stacking or not, specifiable for rnastructure, vfold
    noncanonical(bool): include noncanonical pairs or not (for contrafold, RNAstructure (Cyclefold))
    beam size (int): Beam size for LinearPartition base pair calculation.
    DEBUG (bool): Output command-line calls to packages.
    threshknot (bool): calls threshknot to predict pseudoknots (for contrafold with LinearPartition)

    Possible packages: 'vienna_2', 'vienna_1','contrafold_1','contrafold_2',
    'nupack_95','nupack_99','rnasoft_2007','rnasoft_1999','rnastructure','vfold_0','vfold_1'

    Returns
    array: NxN matrix of base pair probabilities
  '''
    package = package.lower()
    try:
        pkg, version = package.split('_')
    except:
        pkg, version = package, None

    if motif is not None and pkg != 'vienna':
        raise ValueError('motif option can only be used with Vienna.')

    if pseudo and pkg != 'nupack':
        raise ValueError('pseudoknot option only implemented with Nupack.')

    if not dangles and pkg not in ['vienna','nupack']:
        print('Warning: %s does not support dangles options' % pkg)
    if not coaxial and pkg not in ['rnastructure','vfold']:
        print('Warning: %s does not support coaxial options' % pkg)
    if linear and pkg not in ['vienna','contrafold','eternafold']:
        print('Warning: LinearPartition only implemented for vienna, contrafold, eternafold.')

    if pkg=='nupack':
        return bpps_nupack_(sequence, version = version, dangles = dangles, T = T, pseudo=pseudo, dna=dna)

    elif pkg=='vfold':
        return bpps_vfold_(sequence, version = version, T = T, coaxial = coaxial)
    else:

        _, tmp_file = pfunc(sequence, package=package, bpps=True, linear=linear,
            motif=motif, constraint=constraint, T=T, coaxial=coaxial, probing_signal=probing_signal, probing_kws=probing_kws, DIRLOC=package_locs[package],
             dangles=dangles, param_file=param_file,reweight=reweight, beam_size=beam_size, DEBUG=DEBUG, threshknot=threshknot)

        if linear:
            #parse linearpartition output
            return bpps_linearpartition_(sequence, tmp_file)
        else:

            if 'contrafold' in pkg:
                return bpps_contrafold_(sequence, tmp_file)
            if package=='eternafold':
                return bpps_contrafold_(sequence, tmp_file)
            elif 'vienna' in pkg:
                return bpps_vienna_(sequence, tmp_file)
            elif 'rnasoft' in pkg:
                return bpps_rnasoft_(sequence, tmp_file)
            elif 'rnastructure' in pkg:
                return bpps_rnastructure_(sequence, tmp_file, coaxial=coaxial)

            else:
                raise RuntimeError('package not yet implemented')

def bpps_vienna_(sequence, tmp_file):

    dot_fname = tmp_file

    probs=np.zeros([len(sequence), len(sequence)])
    with open(dot_fname,'r') as f:
        for line in f.readlines():
            if 'ubox' in line:
                try:
                    i, j, p, _ = line.split()
                    i, j, p = int(i)-1, int(j)-1, float(p)**2
                    probs[i,j] = p
                    probs[j,i] = p
                except:
                    pass
    os.remove(dot_fname)
    return probs

def bpps_contrafold_(sequence, tmp_file):

    fname = tmp_file

    probs=np.zeros([len(sequence), len(sequence)])

    for line in open(fname).readlines():
        if len(line.split(':')) > 1:
            first_ind = int(line.split()[0])-1
            for x in line.split()[2:]:
                second_ind = int(x.split(':')[0])-1
                p = float(x.split(':')[1])
                probs[first_ind, second_ind] = p
                probs[second_ind, first_ind] = p

    os.remove(fname)

    return probs

def bpps_rnasoft_(sequence, tmp_file):
    fname = tmp_file

    probs=np.zeros([len(sequence), len(sequence)])
    for line in open(fname).readlines():
        i,j,p = int(line.split()[0]), int(line.split()[1]), float(line.split()[2])
        probs[i,j] = p
        probs[j,i] = p

    os.remove(fname)

    return probs

def bpps_nupack_(sequence, version='95', T=37, dangles=True, pseudo=False,dna=False):

    if not version: version='95'

    nupack_materials={'95': 'rna1995', '99': 'rna1999'}

    if dna:
        material='dna1998'
    else:
        material=nupack_materials[version]

    DIR = package_locs['nupack']

    if dangles:
        dangle_option='some'
    else:
        dangle_option='none'

    seqfile = write([sequence])

    command=['%s/pairs' % DIR, '%s' % seqfile.replace('.in',''),
      '-T', str(T), '-material', material, '-dangles', dangle_option, '-cutoff', '0.0000000001']

    if pseudo:
        command.append('--pseudo')
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if p.returncode:
        raise Exception('Nupack pfunc failed: on %s\n%s' % (sequence, stderr))

    ppairs_file = '%s.ppairs' % seqfile.replace('.in','')
    os.remove(seqfile)

    probs=np.zeros([len(sequence), len(sequence)])

    with open(ppairs_file, 'r') as f:
        for line in f.readlines():
            if not line.startswith('%'):
                fields = line.split()
                if len(fields) > 1:
                    if int(fields[1]) <= len(sequence):
                        i, j, p = int(fields[0])-1, int(fields[1])-1, float(fields[2])
                        probs[i,j] = p
                        probs[j,i] = p
    os.remove(ppairs_file)

    return probs

def bpps_rnastructure_(sequence, tmp_file, coaxial=True, DEBUG=False):

    DIR = package_locs['rnastructure']

    pfsfile = tmp_file #'%s/rnastructtmp.pfs' % package_locs['TMP']
    outfile = '%s.probs' % (tmp_file.replace('.pfs',''))
    command = ['%s/ProbabilityPlot' % DIR, pfsfile, outfile, '-t', '-min', '0.0000000001']

    probs=np.zeros([len(sequence), len(sequence)])

    if DEBUG: print(' '.join(command))
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if p.returncode:
        raise Exception('RNAstructure ProbabilityPlot failed: on %s\n%s' % (seq, stderr))

    with open(outfile, 'r') as f:
        for line in f.readlines()[2:]:
            fields = line.split()
            i, j, p = int(fields[0])-1, int(fields[1])-1, 10**(-1*float(fields[2]))
            probs[i,j] = p
            probs[j,i] = p

    os.remove(outfile)
    os.remove(pfsfile)
    return probs

def bpps_vfold_(sequence, version='0',T=37, coaxial=True, DEBUG=False):
    #available versions: 0 for Turner 04 params, 1 for Mfold 2.3 params

    DIR = package_locs["vfold"]

    cwd = os.getcwd()
    os.chdir(DIR) #vfold precompiled binaries don't work being called from elsewhere

    if DEBUG: print(os.getcwd())

    seqfile = write([sequence])

    outfile = filename()+'.pij'

    if sys.platform=="linux":
        platform='linux'
    elif sys.platform=="darwin":
        platform='mac'
    elif sys.platform=="win32":
        platform='win'
    else:
        raise RuntimeError('Vfold has binaries for linux, macOS, and win')

    command = ['./Vfold2d_npk_%s.o %d %d %s %s %d' % (platform, int(coaxial), T, seqfile, outfile, int(version))]

    if DEBUG: print(' '.join(command))

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)

    stdout, stderr = p.communicate()
    os.chdir(cwd)

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('Vfold2d_npk failed: on %s\n%s' % (sequence, stderr))

    os.remove(seqfile)
    probs = np.zeros([len(sequence),len(sequence)])
    p_ij_output = np.loadtxt(outfile,usecols=(0,2,3)) #col 0: set of inds 1, col 1: set of inds 2, col 2: bpp

    for i,j,p in p_ij_output:
    	probs[int(i-1),int(j-1)] = p
    	probs[int(j-1),int(i-1)] = p
    os.remove(outfile)

    return probs
    #output: take second field of last line for Z


def bpps_linearpartition_(sequence, tmp_file):

    fname = tmp_file

    probs=np.zeros([len(sequence), len(sequence)])

    for line in open(fname,'r').readlines():
        if len(line.strip())>0:
            first_ind, second_ind, p = line.strip().split(' ')
            first_ind = int(first_ind)-1
            second_ind = int(second_ind)-1
            p = float(p)
            probs[first_ind, second_ind] = p
            probs[second_ind, first_ind] = p

    os.remove(fname)

    return probs



from arnie.pfunc import pfunc
from arnie.utils import load_package_locations

sample_seq = 'CGCUGUCUGUACUUGUAUCAGUACACUGACGAGUCCCUAAAGGACGAAACAGCG'


def test_pfunc(package):

    Z = pfunc(sample_seq, package=package)
    print('test %s' % package, Z)
    return


if __name__ == '__main__':
    package_locs = load_package_locations()
    for pkg in sorted(package_locs.keys()):

        if (pkg == 'TMP') or (
            pkg.startswith('linear')) or (
            pkg in ['hotknots', 'ipknot', 'knotty', 'pknots', 'spotrna',
                    'spotrna_conda_env', 'e2efold', 'e2efold_conda_env',
                    'spotrna2']):
            print(f'{pkg} not tested.')
            continue
        print(pkg)
        test_pfunc(pkg.lower())




from arnie.sample_structures import sample_structures

sample_seq = 'GGGGAAAACCCC'


def test_sample_seq():

    struct_list = sample_structures(
        sample_seq, n_samples=10, package='vienna_2')
    # sample structures no longer returns energy or prob?
    # print(ener_list) # , ener_list, prob_list
    # print(prob_list)
    return


if __name__ == '__main__':
    test_sample_seq()
    # test_pkg_w_bpps(pkg.lower())



import os, re, sys, shutil
import subprocess as sp
import random, string
import numpy as np
from .utils import *


# load package locations from yaml file, watch! global dict
package_locs = load_package_locations()

def pfunc(seq, package='vienna_2', T=37,
    constraint=None, motif=None, linear=False,
    dangles=True, noncanonical=False, pseudo=False, dna=False, DIRLOC=None,
    bpps=False, param_file=None, coaxial=True, reweight=None,
    return_free_energy = False, beam_size=100, DEBUG=False, threshknot=False,
    probing_signal=None, probing_kws = None):
    ''' Compute partition function for RNA sequence.

        Args:
        seq (str): nucleic acid sequence
        T (float): temperature (Celsius)
        constraint (str): structure constraints
        motif (str): argument to vienna motif
        linear (bool): call LinearPartition to estimate Z in Vienna or Contrafold
        pseudo (bool): nupack only, make prediction with pseudoknots
        dna (bool): nupack only, make prediction for DNA
        dangles (bool): dangles or not, specifiable for vienna, nupack
        coaxial (bool): coaxial stacking or not, specifiable for rnastructure, vfold
        noncanonical(bool): include noncanonical pairs or not (for contrafold, RNAstructure (Cyclefold))
        beam_size (int): beam size option for LinearPartition.
        threshknot (bool): call threshknot to predict pseudoknots (for contrafold, using LinearPartition)

        Possible packages:
        'vienna_2', 'vienna_1','contrafold_1','contrafold_2','nupack_95','nupack_99','rnasoft_2007','rnasoft_1999','rnastructure','vfold_0','vfold_1'

    Returns
        float: free energy
    '''

    if DEBUG: load_package_locations(DEBUG=True)

    try:
        pkg, version = package.lower().split('_')
    except:
        pkg, version = package.lower(), None

    if not bpps: # if bpps, already printed these warnings
        if not dangles and pkg not in ['vienna', 'nupack']:
            print('Warning: %s does not support dangles options' % pkg)
        if not coaxial and pkg not in ['rnastructure', 'vfold']:
            print('Warning: %s does not support coaxial options' % pkg)
        if linear and pkg not in ['vienna','contrafold','eternafold']:
            print('Warning: LinearPartition only implemented for vienna, contrafold, eternafold.')

    # if pkg=='eternafold' and not 'eternafoldparams' in package_locs.keys():
    #     raise RuntimeError('Error: need to set path to EternaFold params to use eternafold hotkey.')


    if pseudo and pkg !='nupack':
        raise ValueError('pseudo only for use with nupack')

    if threshknot:
        if pkg !='contrafold':
            print('Warning: ThreshKnot only implemented for Contrafold')
        elif not linear:
            print('Warning: ThreshKnot only implemented with LinearPartition')

    if pkg=='vienna':
        if linear:
            Z, tmp_file = pfunc_linearpartition_(seq, package='vienna',bpps=bpps, beam_size=beam_size,
                return_free_energy=return_free_energy, DEBUG=DEBUG)

        else:
            Z, tmp_file = pfunc_vienna_(seq, version=version, T=T, dangles=dangles,
             constraint=constraint, motif=motif, bpps=bpps, param_file=param_file,
             reweight=reweight, return_free_energy=return_free_energy, DEBUG=DEBUG, probing_signal=probing_signal, probing_kws = probing_kws,)

    elif pkg=='contrafold':
        if linear:
            Z, tmp_file = pfunc_linearpartition_(seq, package='contrafold', bpps=bpps, beam_size=beam_size,
                return_free_energy=return_free_energy, DEBUG=DEBUG, threshknot=threshknot)
        else:
            Z, tmp_file = pfunc_contrafold_(seq, version=version, T=T,
                constraint=constraint, bpps=bpps, param_file=param_file, DIRLOC=DIRLOC,
                return_free_energy=return_free_energy, DEBUG=DEBUG)

    elif pkg=='rnastructure':
        Z, tmp_file = pfunc_rnastructure_(seq, version=version, T=T, coaxial=coaxial,
            constraint=constraint, bpps=bpps, return_free_energy=return_free_energy, DEBUG=DEBUG)

    elif pkg=='rnasoft':
        if constraint is not None:
            print("ERROR: RNAsoft is unable to handle constraints for calculating \
                partition functions, returning unconstrained Z.")

        Z, tmp_file = pfunc_rnasoft_(seq, version=version, T=T, constraint=constraint,
         bpps=bpps,return_free_energy=return_free_energy, DEBUG=DEBUG)

    elif pkg=='nupack':
        Z, tmp_file = pfunc_nupack_(seq, version=version, dangles=dangles, T=T, pseudo=pseudo, dna=dna, constraint=constraint,

            return_free_energy=return_free_energy, DEBUG=DEBUG)

    elif pkg=='vfold':
        Z, tmp_file = pfunc_vfold_(seq, version=version, T=T, coaxial=coaxial, DEBUG=DEBUG)

    elif pkg=='eternafold':
        if linear:
            Z, tmp_file = pfunc_linearpartition_(seq, package='eternafold', bpps=bpps, beam_size=beam_size,
                return_free_energy=return_free_energy, DEBUG=DEBUG)
        else:
            # Using contrafold code and eternafold params
            if 'eternafoldparams' in package_locs.keys() and 'eternafold' not in package_locs.keys():
                Z, tmp_file = pfunc_contrafold_(seq, T=T, constraint=constraint, 
                    bpps=bpps, param_file=package_locs['eternafoldparams'], DIRLOC=DIRLOC, return_free_energy=return_free_energy, DEBUG=DEBUG)

            elif 'eternafold' in package_locs.keys() and param_file is None:
                #Using eternafold code and params in eternafold codebase
                efold_param_file = package_locs['eternafold'].replace('src','parameters')+'/EternaFoldParams.v1'
                if not os.path.exists(efold_param_file):
                    RuntimeError('Error: Parameters not found at %s' % efold_param_file)

                else:
                    Z, tmp_file = pfunc_contrafold_(seq, T=T, constraint=constraint, 
                        bpps=bpps, param_file=efold_param_file, DIRLOC= package_locs['eternafold'], return_free_energy=return_free_energy, DEBUG=DEBUG)
            elif 'eternafold' in package_locs.keys() and param_file is not None:
                    Z, tmp_file = pfunc_contrafold_(seq, T=T, constraint=constraint, 
                        bpps=bpps, param_file=param_file, DIRLOC= package_locs['eternafold'], 
                        probing_kws=probing_kws, probing_signal = probing_signal, return_free_energy=return_free_energy, DEBUG=DEBUG)                

    else:
        raise ValueError('package %s not understood.' % package)

    if bpps:
        return Z, tmp_file

    else:
        if tmp_file:
            try:
                os.remove(tmp_file)
            except:
                pass
        return Z

def pfunc_vienna_(seq, T=37, version='2', constraint=None, motif=None, param_file=None,
                dangles=True, bpps=False, reweight=None, return_free_energy=False, DEBUG=False, probing_signal=None, shapeMethod='W', probing_kws=None):
    """get partition function structure representation and Z

    Args:
        seq (str): nucleic acid sequence
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif
    Returns
        str, float: secondary structure representation and Z
    """

    if not version:
        version='2'

    if version.startswith('2'):
        LOC=package_locs['vienna_2']
    elif version.startswith('1'):
        LOC=package_locs['vienna_1']
    else:
        raise RuntimeError('Error, vienna version %s not present' % version)

    command = ['%s/RNAfold' % LOC, '-p', '-T', str(T)]


    if probing_signal is not None:
        if probing_kws is None:
            probing_kws={}
        probing_file = run_RNAPVmin(probing_signal, seq, LOC, DEBUG, **probing_kws)
        command.append('--shape=%s' % probing_file)
        command.append('--shapeMethod=%s' % shapeMethod)


    if version.startswith('2'):

        command.append('--bppmThreshold=0.0000000001')
        output_id = local_rand_filename()
        output_dot_ps_file = "%s_0001_dp.ps" % output_id
        command.append('--id-prefix=%s' % output_id)
    else:
        output_dot_ps_file = 'dot.ps'

    if motif is not None:
        command.append("--motif=%s" % motif)

    if constraint is not None:
        fname = write([seq, constraint])
        command.append('-C')
        if version=='2':
            command.append('--enforceConstraint')
    else:
        fname = write([seq])

    if not dangles:
        command.append('--dangles=0')

    if reweight is not None:
        command.append('--commands=%s' % reweight)

    if param_file:
        command.append('--paramFile=%s' % param_file)

    with open(fname) as f:
        if DEBUG: print(fname)
        if DEBUG: print(' '.join(command))
        p = sp.Popen(command, stdin=f, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if stderr.decode('utf-8').startswith('WARNING: '):
        print(stderr)

    if p.returncode:
        raise Exception('RNAfold failed: on %s\n%s' % (seq, stderr))
    os.remove(fname)

    if version.startswith('2'):
        os.remove("%s_0001_ss.ps" % output_id)

    if 'omitting constraint' in stderr.decode('utf-8'):
        free_energy = np.inf # Impossible structure
    else:
        m = re.search('([,|\(\.\)\]\[\{\}]+)\s+\[\s*(-*[0-9]+\.[0-9]+)', stdout.decode('utf-8'))

        free_energy = float(m.group(2))
        if DEBUG: print('free_energy: ', free_energy)

    if return_free_energy:
        return free_energy, output_dot_ps_file
    else: # return Z
        return np.exp(-1*free_energy/(.0019899*(273+T))), output_dot_ps_file

def pfunc_contrafold_(seq, T=37, version='2', constraint=None, bpps=False,
         param_file=None, return_free_energy=False, DIRLOC=None, DEBUG=False, probing_signal=None, probing_kws=None):

    """get partition function structure representation and free energy

    Args:
        seq (str): nucleic acid sequence
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif

        DIRLOC: sets location of contrafold specifically (Useful if there's several EternaFold builds to compare.)
    Returns
        float: partition function
        Note: If the constraint is impossible then Z wil be equal to the Z unconstrained
    """
    if not version: version='2'

    if probing_signal is not None:
        fname = write_reactivity_file_contrafold(probing_signal, seq)
    else:
        fname = '%s.in' % filename()

    if DIRLOC is not None:
        LOC=DIRLOC
    elif version.startswith('2'):
        LOC=package_locs['contrafold_2']
    elif version.startswith('1'):
        LOC=package_locs['contrafold_1']
    else:
        raise RuntimeError('Error, Contrafold version %s not present' % version)

    command = ['%s/contrafold' % LOC, 'predict', fname]

    if probing_signal is not None:
        command = command + ['--evidence', '--params', package_locs['eternafoldparams_PLUS_POTENTIALS'], '--numdatasources','1', ]
        if probing_kws is not None:
            if 'kappa' in probing_kws.keys():
                command = command + ['--kappa', str(probing_kws['kappa']) ]
    else:
        if param_file is not None:
            command = command + ['--params', param_file]

    if bpps:
        posterior_fname = '%s.posteriors' % filename()
        command = command + ['--posteriors', '0.0000000001', posterior_fname]

    else:
        command.append('--partition')


    if constraint is not None:
        convert_dbn_to_contrafold_input(seq, constraint, fname)
        command.append('--constraints')
    else:
        if probing_signal is None:
            convert_dbn_to_contrafold_input(seq, ''.join(['.' for x in range(len(seq))]), fname)

    if DEBUG: print(' '.join(command))
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('Contrafold failed: on %s\n%s' % (seq, stderr))

    os.remove(fname)

    if not bpps:
        logZ = float(stdout.decode('utf-8').rstrip().split()[-1])

        if return_free_energy:
            return -1*logZ, None
        else:
            return np.exp(logZ), None
    else:
        return 0, posterior_fname

def pfunc_rnasoft_(seq, version='99', T=37, constraint=None, bpps=False, return_free_energy=False, DEBUG=False):
    DIR = package_locs['rnasoft']

    if not version: version='blstar'

    #note for mfe will use simfold instead of simfold pf

    # supported versions: 07, 99, 99-no-dangles, BL-no-dangles, BLstar, LAM-CG, NOM-CG
    param_locs = {'07': '%s/params/CG_best_parameters_ISMB2007.txt' % DIR,
    '99': '%s/params/turner_parameters_fm363_constrdangles.txt' % DIR,
    '99-no-dangles': '%s/params/turner_parameters_fm363_dangles0.txt' % DIR,
    'bl-no-dangles': '%s/params/BL-no-dangles.txt' % DIR,
    'blstar': '%s/params/BLstar.txt' % DIR,
    'lam-cg': '%s/params/LAM-CG.txt' % DIR,
    'nom-cg': '%s/params/NOM-CG.txt' % DIR}

    command = ['%s/simfold_pf' % DIR, '-s', seq, '-p', param_locs[version]]

    if DEBUG: print(' '.join(command))
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    bpps_fname = '%s.bpps' % filename()
    if bpps:
        with open(bpps_fname,'w') as f:
            for line in stdout.decode('utf-8').split('\n')[5:]:
                if not 'Glog' in line and len(line) > 1:
                    f.write(line+'\n')

    if p.returncode:
        raise Exception('RNAsoft partition failed: on %s\n%s' % (seq, stderr))

    Z = float(stdout.decode('utf-8').split('\n')[1].split()[-1])

    if return_free_energy:
        return -1*np.log(Z), bpps_fname
    else:
        return Z, bpps_fname

def pfunc_nupack_(seq, version='95', T=37, dangles=True, constraint=None, return_free_energy=False, pseudo=False, dna=False, DEBUG=False):

    if not version: version='95'
    nupack_materials={'95': 'rna1995', '99': 'rna1999', 'dna':'dna1998'}

    if dna: version='dna'

    DIR = package_locs['nupack']

    if dangles:
        dangle_option='some'
    else:
        dangle_option='none'

    if constraint is not None:
        if '.' in constraint:
            print('Warning: NUPACK does not do constrained folding . and x')
        seqfile = write([seq, constraint])
        command=['%s/energy' % DIR, '%s' % seqfile.replace('.in',''),'-T', str(T),
             '-material', nupack_materials[version], '-dangles', dangle_option]

    else:
        seqfile = write([seq])
        command=['%s/pfunc' % DIR, '%s' % seqfile.replace('.in',''),'-T', str(T),
             '-material', nupack_materials[version], '-dangles', dangle_option]

    if pseudo:
        command.append('--pseudo')
    if DEBUG: print(' '.join(command))
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if p.returncode:
        raise Exception('Nupack pfunc failed: on %s\n%s' % (seq, stderr))

    if constraint is not None:
        free_energy = float(stdout.decode('utf-8').split('\n')[-2])
        Z=np.exp(-1*free_energy/(.0019899*(273+T)))

    else:
        free_energy = float(stdout.decode('utf-8').split('\n')[-3])
        Z=float(stdout.decode('utf-8').split('\n')[-2])

    os.remove(seqfile)

    if return_free_energy:
        return free_energy, None
    else:
        return Z, None

def pfunc_rnastructure_(seq, version=None, T=37, constraint=None, coaxial=True,
                            bpps=False, return_free_energy=False, DEBUG=False):
    """get partition function structure representation and free energy

    Args:
        seq (str): nucleic acid sequence
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif
        coaxial (bool): Coaxial stacking or not (default True)
    Returns
        float: partition function
    """

    seqfile = write([seq])
    pfsfile = '%s.pfs' % filename()
    DIR = package_locs['rnastructure']
    command = ['%s/partition' % DIR, seqfile, pfsfile, '-T', str(T+273)]

    if not coaxial:
        command.extend(['--disablecoax'])

    if constraint is not None:
        fname = '%s.CON' % filename()
        #print(fname)
        convert_dbn_to_RNAstructure_input(seq, constraint, fname)
        command.extend(['--constraint', fname])

    if DEBUG: print(' '.join(command))
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    os.remove(seqfile)
    if constraint is not None:
        os.remove(fname)
    
    if p.returncode:
        raise Exception('RNAstructure partition failed: on %s\n%s' % (seq, stderr))

    if not bpps:
        command = ['%s/EnsembleEnergy' % DIR, pfsfile]

        if DEBUG: print(' '.join(command))
        p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

        stdout, stderr = p.communicate()

        if DEBUG:
            print('stdout')
            print(stdout)
            print('stderr')
            print(stderr)

        if p.returncode:
            raise Exception('RNAstructure EnsembleEnergy failed: on %s\n%s' % (seq, stderr))

        if DEBUG: print(stdout.decode('utf-8').split('\n')[3])
        free_energy = float(stdout.decode('utf-8').split('\n')[3].split(' ')[-2])

        if return_free_energy:
            return free_energy, pfsfile
        else:
            return np.exp(-1*free_energy/(.0019*(273+T))), pfsfile
    else:
        return 0, pfsfile

def pfunc_vfold_(seq, version='0', T=37, coaxial=True, bpps=False, DEBUG=False):
    #available versions: 0 for Turner 04 params, 1 for Mfold 2.3 params
    #for bpps
        # command = ['%s/Vfold2d_npk_mac.o %d %d %s %s %d' % (DIR, int(coaxial),\
        #  T, infile, outfile, int(version))]

    DIR = package_locs["vfold"]

    cwd = os.getcwd()
    os.chdir(DIR) #vfold precompiled binaries don't work being called from elsewhere

    if DEBUG: print(os.getcwd())

    seqfile = write([seq])

    if sys.platform=="linux":
        platform='linux'
    elif sys.platform=="darwin":
        platform='mac'
    elif sys.platform=="win32":
        platform='win'
    else:
        raise RuntimeError('Vfold has binaries for linux, macOS, and win')


    command = ['./VfoldThermal_npk_%s.o %d %d %d %s tmp %d; cat tmp; rm tmp' % (platform, int(coaxial), T, T, seqfile, int(version))]

    if DEBUG: print(' '.join(command))

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)

    stdout, stderr = p.communicate()
    os.chdir(cwd)

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('VfoldThermal_npk failed: on %s\n%s' % (seq, stderr))

    Z=float(stdout.decode('utf-8').split('\n')[-2].split()[1])

    os.remove(seqfile)
    return Z, None
    #output: take second field of last line for Z


def pfunc_linearpartition_(seq, bpps=False, package='contrafold', beam_size=100, return_free_energy=False, DEBUG=False, threshknot=False):
    LOC = package_locs['linearpartition']
    tmp_file = filename()
    tmp_command = filename()

    if bpps:
        pf_only = 0
    else:
        pf_only = 1

    if threshknot:
        TK = '-T'
    else:
        TK = '_'

    # args: beamsize, is_sharpturn, is_verbose, bpp_file, bpp_prefix, pf_only, bpp_cutoff,
    #forest_file, mea, gamma, TK, threshold, ThreshKnot_prefix, MEA_prefix, MEA_bpseq

    # args: beamsize, is_sharpturn, is_verbose, bpp_file, bpp_prefix, pf_only, bpp_cutoff,
    #forest_file, mea, gamma, TK, threshold, ThreshKnot_prefix, MEA_prefix, MEA_bpseq, shape_file_path

    #threshknot threshold set to default 0.3

    command=['echo %s | %s/linearpartition_%s' % (seq, LOC, package[0].lower()), str(beam_size),
     '0', '0', tmp_file, "''", str(pf_only), '0.000001', "''", "''", "''",'%s' % (TK), "''", "''", "''", "''", "''"]

    with open('%s.sh' % tmp_command,'w') as f:
        f.write(' '.join(command))

    if DEBUG: print(' '.join(command))

    meta_command = ['chmod +x %s.sh; %s.sh' % (tmp_command, tmp_command)]
    p = sp.Popen(meta_command, stdout=sp.PIPE, stderr=sp.PIPE,shell=True)

    stdout, stderr = p.communicate(input=str.encode(seq))

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if p.returncode:
        raise Exception('LinearPartition failed: on %s\n%s' % (seq, stderr))

    os.remove("%s.sh" % tmp_command)
    # Note: the linearfold exec says this is free energy in kcal/mol.
    # this is still just cfold log Z

    # linearfold returns two different things depending on which package

    if bpps:
        return 0, tmp_file
    else:

        if package in ['contrafold','eternafold']:
            logZ=float(stderr.decode('utf-8').split(' ')[-1])
            #os.remove('_') # tmp file written by linearpartition forest
            if return_free_energy:
                return -1*logZ, None
            else:
                return np.exp(logZ), None

        elif package=='vienna':
            free_energy = float(stderr.decode('utf-8').split(' ')[-2])
            T=37
            #os.remove('_') # tmp file written by linearpartition forest
            if return_free_energy:
                return free_energy, None
            else:
                return np.exp(-1*free_energy/(.0019899*(273+T))), None



from arnie.bpps import bpps
from arnie.utils import load_package_locations

sample_seq = 'CGCUGUCUGUACUUGUAUCAGUACACUGACGAGUCCCUAAAGGACGAAACAGCG'


def test_bpps(pkg):
    p = bpps(sample_seq, package=pkg)
    print('test bpps %s' % pkg)
    print(p[0])


if __name__ == '__main__':
    print("Test: printing first row of bpp matrices")
    package_locs = load_package_locations()
    PK_packages = ['hotknots', 'ipknot', 'knotty', 'pknots',
                   'spotrna', 'spotrna_conda_env', 'e2efold',
                   'e2efold_conda_env', 'spotrna2']
    for pkg in sorted(package_locs.keys()):
        if pkg == 'TMP' or pkg.startswith('linear') or pkg in PK_packages:
            continue

        test_bpps(pkg)




import os
from arnie.utils import load_package_locations


def test_settings():
    package_locs = load_package_locations()
    for k in package_locs.keys():
        print(k)
        assert os.path.isdir(package_locs[k])
    return


if __name__ == '__main__':
    test_settings()



from arnie import pfunc

# 3MXH c-di-GMP riboswitch, has coaxial stacking
sample_seq = 'GGUCACGCACAGGGCAAACCAUUCGAAAGAGUGGGACGCAAAGCCUCCGGCCUAAACCAUUGCACUCCGGUAGGUAGCGGGGUUACCGAUGG'


def test_pkg(package, coaxial=True):

    Z = pfunc.pfunc(sample_seq, package=package, bpps=False, coaxial=coaxial)
    print('test %s' % package, Z)
    return None


if __name__ == '__main__':
    for pkg in ['vfold_0', 'vfold_1']:
        for coaxial in [True, False]:
            print(pkg, "coaxial %d" % coaxial)
            test_pkg(pkg, coaxial=coaxial)



import subprocess as sp
from arnie.utils import *
import glob
from os import getcwd, chdir, remove, mkdir, rmdir, path
from scipy.optimize import linear_sum_assignment


# TODO script all previous investigations
# TODO Debug modes to print output and err to help with install issues
# TODO pk_predict options +

package_locs = load_package_locations()


def pk_predict(seq, predictor,
               model="default", param="parameters_DP03.txt",
               refinement=1, t1="auto", t2='auto',
               cpu=32):
    '''

    ipknot options:
        model: one of ["LinearPartition-C","LinearPartition-V","Boltzmann","ViennaRNA","CONTRAfold","NUPACK"]
        t1: probability threshold level 1 
        t2: probability threshold level 2
        refinement: number of times for refinment

    hotknots options:
        model: one of ["CC","RE","DP"]
        param: one of ["parameters_CC06.txt","parameters_CC09.txt","parameters_DP03.txt","parameters_DP09.txt"]

    spotrna options:
        cpu: number cpu threads

    e2efold options:
        ???

    nupack options:
        ????

    '''
    if predictor not in ["hotknots", "ipknot", "knotty", "spotrna", "e2efold", "pknots","spotrna2","nupack"]:
        raise ValueError('Only hotknots,ipknot,knotty,spotrna,spotrna2,e2efold,pknots,nupack implemented.')
    if predictor == "spotrna":
        return _run_spotrna(seq, cpu=cpu)[0]
    elif predictor == "spotrna2":
        return _run_spotrna2(seq)[0]
    elif predictor == "e2efold":
        return _e2efold(seq)
    elif predictor == "pknots":
        return _pknots(seq)
    elif predictor == "knotty":
        return _knotty_mfe(seq)
    elif predictor == "hotknots":
        if model == "default":
            model = "DP"
        if model not in ["CC", "RE", "DP"]:
            raise ValueError('Only CC, RE, DP model implemented for hotknots.')
        if param not in ["parameters_CC06.txt", "parameters_CC09.txt", "parameters_DP03.txt", "parameters_DP09.txt"]:
            raise ValueError('Only parameters_CC06.txt, parameters_CC09.txt, parameters_DP03.txt, parameters_DP09.txt parameters implemented for hotknots.')
        return _run_hotknots(seq, model=model, param=param)[0][0]
    elif predictor == "ipknot":
        if model == "default":
            model = "LinearPartition-C"
        if model not in ["LinearPartition-C", "LinearPartition-V", "Boltzmann", "ViennaRNA", "CONTRAfold", "NUPACK"]:
            raise ValueError('Only LinearPartition-C, LinearPartition-V, Boltzmann, ViennaRNA, CONTRAfold, NUPACK model implemented for ipknot.')
        return _ipknot_mfe(seq, model=model, refinement=refinement, t1=t1, t2=t2)
    elif predictor == "nupack":
        return _nupack_mfe_pk(seq)


def pk_predict_from_bpp(bpp, heuristic="hungarian", theta=None, allowed_buldge_len=0, min_len_helix=2,
                        exp=1, sigmoid_slope_factor=None, prob_to_0_threshold_prior=0, prob_to_1_threshold_prior=1, ln=False, add_p_unpaired=True,
                        max_iter=1):
    '''
    threshknot options:
        theta
        max_iter
        allowed_buldge_len
        min_len_helix

    hungarian options:
        add_p_unpaired
        theta (aka prob_to_0_threshold_post)
        prob_to_0_threshold_prior
        prob_to_1_threshold_prior
        exp
        sigmoid_slope_factor
        ln
        allowed_buldge_len
        min_len_helix
    '''

    if heuristic not in ["threshknot", "hungarian"]:
        raise ValueError('Only threshknot and hunagrian heuristics implemented.')

    if heuristic == "threshknot":
        if theta is None:
            theta = 0.3
        return _threshknot(bpp, theta=theta, max_iter=max_iter, allowed_buldge_len=allowed_buldge_len, min_len_helix=min_len_helix)[0]
    elif heuristic == "hungarian":
        if theta is None:
            theta = 0.0
        return _hungarian(bpp, exp=1, sigmoid_slope_factor=sigmoid_slope_factor, prob_to_0_threshold_prior=prob_to_0_threshold_prior,
                          prob_to_1_threshold_prior=prob_to_1_threshold_prior, theta=theta, ln=ln, add_p_unpaired=add_p_unpaired,
                          allowed_buldge_len=allowed_buldge_len, min_len_helix=min_len_helix)[0]


def _hungarian(bpp, exp=1, sigmoid_slope_factor=None, prob_to_0_threshold_prior=0,
               prob_to_1_threshold_prior=1, theta=0, ln=False, add_p_unpaired=True,
               allowed_buldge_len=0, min_len_helix=2):

    bpp_orig = bpp.copy()

    if add_p_unpaired:
        p_unpaired = 1 - np.sum(bpp, axis=0)
        for i, punp in enumerate(p_unpaired):
            bpp[i, i] = punp

    # apply prob_to_0 threshold and prob_to_1 threshold
    bpp = np.where(bpp < prob_to_0_threshold_prior, 0, bpp)
    bpp = np.where(bpp > prob_to_1_threshold_prior, 1, bpp)

    # aply exponential. On second thought this is likely not as helpful as sigmoid since
    # * for 0 < exp < 1 lower probs will increase more than higher ones (seems undesirable)
    # * for exp > 1 all probs will decrease, which seems undesirable (but at least lower probs decrease more than higher ones)
    bpp = np.power(bpp, exp)

    # apply log which follows botlzamann where -ln(P) porportional to Energy
    if ln:
        bpp = np.log(bpp)

    bpp = np.where(np.isneginf(bpp), -1e10, bpp)
    bpp = np.where(np.isposinf(bpp), 1e10, bpp)

    # apply sigmoid modified by slope factor
    if sigmoid_slope_factor is not None and np.any(bpp):
        bpp = _sigmoid(bpp, slope_factor=sigmoid_slope_factor)

        # should think about order of above functions and possibly normalize again here

        # run hungarian algorithm to find base pairs
    _, row_pairs = linear_sum_assignment(-bpp)
    bp_list = []
    for col, row in enumerate(row_pairs):
        # if bpp_orig[col, row] != bpp[col, row]:
        #    print(col, row, bpp_orig[col, row], bpp[col, row])
        if bpp_orig[col, row] > theta and col < row:
            bp_list.append([col, row])

    structure = convert_bp_list_to_dotbracket(bp_list, bpp.shape[0])
    structure = post_process_struct(structure, allowed_buldge_len, min_len_helix)
    bp_list = convert_dotbracket_to_bp_list(structure, allow_pseudoknots=True)

    return structure, bp_list


def _sigmoid(x, slope_factor=0.5):
    # normalize to [-1, 1]
    numerator = (x - x.min()) * 2.0
    denominator = x.max() - x.min()
    #print(numerator, denominator)
    x = numerator / (denominator + 1e-6) - 1.0
    return 1 / (1 + np.exp(-x / slope_factor))


def _threshknot(bpp, theta=0.3, max_iter=1, allowed_buldge_len=0, min_len_helix=2):
    iteration = 0
    length = bpp.shape[0]
    bp_list = []
    new_bp = 1
    while new_bp != 0 and iteration <= max_iter:
        current_bp_list = []
        bp_list_flat = np.array(bp_list).flatten()
        if np.any(bp_list_flat):
            bpp_update = np.delete(bpp, bp_list_flat, axis=1)
            if np.any(bpp_update):
                Pmax = np.amax(bpp_update, axis=1)
        else:
            Pmax = np.amax(bpp, axis=1)
        for i in range(length):
            for j in range(i + 1, length):
                if i not in bp_list_flat and j not in bp_list_flat:
                    prob = bpp[i, j]
                    if prob == Pmax[i] and prob == Pmax[j] and prob > theta:
                        current_bp_list.append([i, j])
        new_bp = len(current_bp_list)
        iteration += 1
        if new_bp != 0 and iteration > max_iter:
            print("Reached max iteration, stopping before converged.")
        else:
            bp_list.extend(current_bp_list)

    bp_list = _check_bp_list(bp_list)
    structure = convert_bp_list_to_dotbracket(bp_list, length)
    structure = post_process_struct(structure, allowed_buldge_len, min_len_helix)
    bp_list = convert_dotbracket_to_bp_list(structure, allow_pseudoknots=True)
    return structure, bp_list


def _check_bp_list(bp_list):
    for bp in bp_list:
        bp.sort()
    bp_list.sort(key=lambda x: x[0])
    nts = [nt for bp in bp_list for nt in bp]
    if len(nts) > len(set(nts)):
        print("WARNING some nucletotides found in more than 1 bp")
        for i, bpA in enumerate(bp_list):
            for bpB in bp_list[i + 1:]:
                if bpA[0] == bpB[0] and bpA[1] == bpB[1]:
                    print("removing repeat bp", bpA)
                    bp_list = bp_list[:i] + bp_list[i + 1:]
                elif bpA[0] in bpB:
                    if abs(bpA[0] - bpA[1]) <= abs(bpB[0] - bpB[1]):
                        to_remove = bpB
                    else:
                        to_remove = bpA
                    print("WARNING base", bpA[0], "is in 2 basepairs", bpA, bpB, "THIS SHOULD BE FIXED. Removing", to_remove)
                    bp_list.remove(to_remove)
                elif bpA[1] in bpB:
                    if abs(bpA[0] - bpA[1]) <= abs(bpB[0] - bpB[1]):
                        to_remove = bpB
                    else:
                        to_remove = bpA
                    print("WARNING base", bpA[1], "is in 2 basepairs", bpA, bpB, "THIS SHOULD BE FIXED. Removing", to_remove)
                    bp_list.remove(to_remove)
    return bp_list


def _run_hotknots(seq, model="DP", param="parameters_DP03.txt"):
    hotknot_location = package_locs["hotknots"]
    cur_dir = getcwd()
    chdir(hotknot_location)
    command = [f"{hotknot_location}/HotKnots", "-noPS", "-s", seq, "-m", model, "-p", f"{hotknot_location}/params/{param}"]
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = p.communicate()
    if p.returncode:
        print('ERROR: hotknots failed: on %s\n%s\n%s' % (seq, out.decode(), err.decode()))
        return ["x"*len(seq)]
    output = out.decode().split("\n")[2:-1]
    structs = []
    for struct in output:
        x = struct.split('\t')
        x2 = [x[0].split(" ")[-1], x[1]]
        structs.append(x2)
    chdir(cur_dir)
    return structs


def _ipknot_mfe(seq, model="LinearPartition-C", refinement=1, t1="auto", t2="auto"):
    """
    TODO
      -g, --gamma G             The weight for true base-pairs equivalent to 
                                '-t 1/(gamma+1)'
      -i, --allow-isolated      Allow isolated base-pairs
      -P, --param FILE          Read the energy parameter file for Vienna RNA 
                                package
      -x, --aux                 Import an auxiliary file for base-pairing 
                                probabilities
      -u, --no-levelwise        Do not perform the levelwise prediction
      -E, --energy              Output with the free energy
    """
    ipknot_location = package_locs["ipknot"]
    out_folder = get_random_folder()
    mkdir(out_folder)
    fasta_file = f"{out_folder}/temp.fasta"
    f = open(fasta_file, "w")
    f.write(">seq \n")
    f.write(seq)
    f.close()
    command = [f"{ipknot_location}/ipknot", fasta_file, "--model", model, "-r", str(refinement), "-t", str(t1), "-t", str(t2)]
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = p.communicate()
    if p.returncode:
        print('ERROR: ipknot failed: on %s\n%s\n%s' % (seq, out.decode(), err.decode()))
        remove(fasta_file)
        rmdir(out_folder)
        return "x"*len(seq)
    output = out.decode().split("\n")
    remove(fasta_file)
    rmdir(out_folder)
    return output[2]


def _knotty_mfe(seq):
    knotty_location = package_locs["knotty"]
    command = [f"{knotty_location}/knotty", seq]
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE,universal_newlines=True)
    try:
        out, err = p.communicate()
    except:
        print("ERROR knotty, could not communicate")
        return "x"*len(seq)
    if p.returncode:
        print('ERROR: knotty failed: on %s\n%s\n%s' % (seq, out, err))
        return "x"*len(seq)
    output = out.split("\n")
    struct = output[1].split(" ")[1]
    bp_list = convert_dotbracket_to_bp_list(struct, allow_pseudoknots=True)
    struct = convert_bp_list_to_dotbracket(bp_list, seq_len=len(struct))
    return struct


def _run_spotrna(seq, cpu=32):
    '''
    SPOT-RNA
    '''
    spotrna_location = package_locs["spotrna"]
    spotrna_conda_env = package_locs["spotrna_conda_env"]
    out_folder = get_random_folder()
    mkdir(out_folder)
    fasta_file = f"{out_folder}/temp.fasta"
    f = open(fasta_file, "w")
    f.write(">seq\n")
    f.write(seq)
    f.close()
    command = [f"{spotrna_conda_env}/python3", f"{spotrna_location}/SPOT-RNA.py", "--inputs", fasta_file, "--outputs", out_folder, "--cpu", str(cpu)]
    # keep running until output file exists
    while not path.exists(out_folder + "/seq.bpseq"):
        p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
        out, err = p.communicate()
        # print(seq, out.decode(),err.decode())
        if p.returncode:
            print('ERROR: spotrna failed: on %s\n%s\n%s' % (seq, out.decode(), err.decode()))
            return "x"*len(seq)
    bp_list = bpseq_to_bp_list(out_folder + "/seq.bpseq")
    struct = convert_bp_list_to_dotbracket(bp_list, len(seq))
    bpp = prob_to_bpp(out_folder + "/seq.prob")
    remove(out_folder + "/seq.bpseq")
    remove(out_folder + "/seq.prob")
    remove(out_folder + "/seq.ct")
    remove(fasta_file)
    rmdir(out_folder)
    return struct, bpp

def _run_spotrna2(seq):
    # TODO 
    spotrna2_location = package_locs["spotrna2"]
    out_folder = get_random_folder()
    mkdir(out_folder)
    fasta_file = f"{out_folder}/temp.fasta"
    f = open(fasta_file, "w")
    f.write(">seq\n")
    f.write(seq)
    f.close()
    command = [f"{spotrna2_location}/run_spotrna2.sh", fasta_file]
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = p.communicate()
    if p.returncode:
        print('ERROR: spotrna2 failed: on %s\n%s\n%s' % (seq, out.decode(), err.decode()))
        return "x"*len(seq)
    bp_list = bpseq_to_bp_list(f"{out_folder}/temp_outputs/temp.bpseq")
    struct = convert_bp_list_to_dotbracket(bp_list, len(seq))
    bpp = prob_to_bpp(f"{out_folder}/temp_outputs/temp.prob")
    for f in os.listdir(f"{out_folder}/temp_outputs"):
        remove(f)
    rmdir(f"{out_folder}/temp_outputs")
    for f in os.listdir(f"{out_folder}/temp_features"):
        remove(f)
    rmdir(f"{out_folder}/temp_features")
    remove(fasta_file)
    rmdir(out_folder)
    return struct, bpp

def _e2efold(seq):
    # only if <600
    # TODO probably plenty of options
    e2efold_location = package_locs["e2efold"]
    e2efold_conda_env = package_locs["e2efold_conda_env"]
    out_folder = get_random_folder()
    mkdir(out_folder)
    with open(f'{out_folder}/config.json', 'w') as f:
        f.write('\n'.join(['{',
                           '        "exp_name": "performance on short sequences (50-600)",',
                           f'        "test_folder": "{out_folder}/short_seqs",',
                           f'        "save_folder": "{out_folder}/short_cts",',
                           '        "gpu": "0",',
                           '        "u_net_d": 10,',
                           '        "BATCH_SIZE": 8,',
                           '        "batch_size_stage_1": 20,',
                           '        "batch_size_stage_2": 16,',
                           '        "OUT_STEP": 100,',
                           '        "LOAD_MODEL": true,',
                           '        "pp_steps": 20,',
                           '        "pp_loss": "f1",',
                           '        "pp_model": "mixed",',
                           '        "rho_per_position": "matrix",',
                           '        "data_type": "rnastralign_all_600",',
                           '        "model_type": "att_simple_fix",',
                           '        "epoches_first": 50,',
                           '        "epoches_second": 10,',
                           '        "epoches_third": 10,',
                           '        "evaluate_epi": 1,',
                           '        "evaluate_epi_stage_1": 5,',
                           '        "step_gamma": 1,',
                           '        "k": 1,',
                           '        "test": {',
                           '                "f1": true,',
                           '                "accuracy": false,',
                           '                "energy": false',
                           '        }',
                           '}']))
    mkdir(f'{out_folder}/short_seqs')
    mkdir(f'{out_folder}/short_cts')
    command = [f"{e2efold_conda_env}/python", f"{e2efold_location}/e2efold_productive_short.py", "-c", f"{out_folder}/config.json"]
    fasta_file = f"{out_folder}/short_seqs/temp.seq"
    f = open(fasta_file, "w")
    f.write(seq)
    f.close()
    # keep running until output file exists
    while not path.exists(f"{out_folder}/short_cts/temp.seq.ct"):
        out, err = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE).communicate()
    bp_list = ct_to_bp_list(f"{out_folder}/short_cts/temp.seq.ct", 1)
    struct = convert_bp_list_to_dotbracket(bp_list, len(seq))
    remove(fasta_file)
    remove(f"{out_folder}/short_cts/temp.seq.ct")
    remove(f"{out_folder}/config.json")
    rmdir(f'{out_folder}/short_seqs')
    rmdir(f'{out_folder}/short_cts')
    rmdir(out_folder)
    return struct


def _pknots(seq):
    ''' TODO
      -a          : pseudoknot approx, exclude V7-V10 and WB9-WB1
      -c          : add L^5 coaxials (V6)
      -s          : shuffle sequences
    '''
    pknots_location = package_locs["pknots"]
    out_folder = get_random_folder()
    mkdir(out_folder)
    fasta_file = f"{out_folder}/temp.fasta"
    f = open(fasta_file, "w")
    f.write(">seq \n")
    f.write(seq)
    f.close()
    outfile = f"{out_folder}/out.out"
    command = [pknots_location + "/pknots", "-k", "-g", fasta_file, outfile]
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = p.communicate()
    remove(fasta_file)
    if p.returncode:
        print('ERROR: PKNOTS failed: on %s\n%s\n%s' % (seq, out.decode(), err.decode()))
        return "x"*len(seq)
    bp_list = ct_to_bp_list(outfile, 4)
    remove(outfile)
    rmdir(out_folder)
    struct = convert_bp_list_to_dotbracket(bp_list, len(seq))
    return struct


def _nupack_mfe_pk(seq):
    # TODO many nupack options... also why is this not implemented in mfe?
    nupack_location = package_locs['nupack']
    out_folder = get_random_folder()
    mkdir(out_folder)
    fasta_file = f"{out_folder}/temp"
    f = open(f'{fasta_file}.in','w')
    f.write(seq)
    f.close()
    struct = None
    command = [nupack_location+'/mfe', "-pseudo", fasta_file]
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)
    out,err = p.communicate()
    if p.returncode:
        print(f'ERROR: nupack mfe pk failed on {seq} {fasta_file} {out.decode} {err.decode}')
        return 'x'*len(seq)
    f = open(f'{fasta_file}.mfe')
    struct = f.readlines()[16][:-1]
    f.close()
    remove(f'{fasta_file}.in')
    remove(f'{fasta_file}.mfe')
    rmdir(out_folder)
    return struct




import os, re, sys, shutil
import subprocess as sp
import random, string
import numpy as np
from .utils import *

DEBUG=False

# load package locations from yaml file, watch! global dict
package_locs = load_package_locations()

def sample_structures(seq, n_samples = 10, package='vienna_2', T=37, constraint=None, param_file=None,
	dangles=True, reweight=None, nonredundant=False):
    ''' Draw stochastic sampled structures for RNA sequence. Possible packages: 'eternafold', 'vienna_2'

        Args:
        seq (str): nucleic acid sequence
        T (float): temperature (Celsius)
        constraint (str): structure constraints
        motif (str): argument to vienna motif 
        dangles (bool): dangles or not, specifiable for vienna, nupack
        noncanonical(bool): include noncanonical pairs or not (for contrafold, RNAstructure (Cyclefold))
        
    Returns
        list of structures
        list of energies
        list of probabilities 
    '''

    try:
        pkg, version = package.lower().split('_')
    except:
        pkg, version = package.lower(), None

    if not dangles and pkg not in ['vienna','nupack']:
        print('Warning: %s does not support dangles options' % pkg)

    if pkg=='vienna':
        struct_list = sample_vienna_(seq, n_samples=n_samples, version=version, T=T, 
        	dangles=dangles, constraint=constraint, reweight=reweight, nonredundant = nonredundant)

    elif pkg=='eternafold':
        struct_list = sample_eternafold_(seq, n_samples=n_samples, param_file=param_file, constraint=constraint, nonredundant = nonredundant)

    else:
        raise ValueError('package %s either not understood or not supported at this moment.' % package)

    return struct_list

def sample_vienna_(seq, n_samples=10, T=37, version='2', constraint=None, 
	dangles=True, reweight=None, nonredundant=False):
    """Stochastically sample structures from Vienna RNAsubopt.

    Inputs:
        seq (str): nucleic acid sequence
        n_samples (int): number of structures to sample.
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif  
    Outputs:
        struct_list (list): list of stochastically-sampled structures.
    """

    if not version:
        version='2'

    if version.startswith('2'):
        LOC=package_locs['vienna_2']
    elif version.startswith('1'):
        LOC=package_locs['vienna_1']
    else:
        raise RuntimeError('Error, vienna version %s not present' % version)

    command = ['%s/RNAsubopt' % LOC, '-T', str(T), '--stochBT_en=%d' % n_samples]#, '-N']

    if constraint is not None:
        fname = write([seq, constraint])
        command.append('-C')
        #command.append('--enforceConstraint')
    else:
        fname = write([seq])

    if not dangles:
        command.append('--dangles=0')

    if nonredundant:
    	command.append('-N')

    if reweight is not None:
        command.append('--commands=%s' % reweight)

    with open(fname) as f:
        if DEBUG: print(fname)
        if DEBUG: print(' '.join(command))
        p = sp.Popen(command, stdin=f, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if p.returncode:
        raise Exception('RNAsubopt failed: on %s\n%s' % (seq, stderr))
    os.remove(fname)

    if 'omitting constraint' in stderr.decode('utf-8'):
        raise RuntimeError("Constraint omitted, Impossible structure")

    else:
        struct_list, prob_list, energy_list = [],[],[]
        output_lines = stdout.decode('utf-8').split('\n')[1:-1] # first line is just repeating sequence, last is empty space
        for line in output_lines:
            struct_list.append(line.split(' ')[0])
            # prob_list.append(float(line.split(' ')[-2]))
            # energy_list.append(float(line.split(' ')[-1]))

    return struct_list

def sample_eternafold_(seq, n_samples=10, param_file=None, constraint=None, nonredundant=False):
    """Stochastically sample structures from EternaFold.

    Inputs:
        seq (str): nucleic acid sequence
        n_samples (int): number of structures to sample.
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif  
    Outputs:
        struct_list (list): list of stochastically-sampled structures.
    """

    fname = '%s.in' % filename()
    LOC=package_locs['eternafold']


    command = ['%s/contrafold' % LOC, 'sample', fname]

    if param_file is not None:
        command = command + ['--params', param_file]
    else:
        command = command + ['--params', package_locs['eternafoldparams']]

    if constraint is not None:
        convert_dbn_to_contrafold_input(seq, constraint, fname)
        command.append('--constraints')
    else:
        convert_dbn_to_contrafold_input(seq, ''.join(['.' for x in range(len(seq))]), fname)

    if DEBUG: print(' '.join(command))

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    struct_list = stdout.decode('utf-8').split('\n')[:-1]

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('Eternafold sample failed: on %s\n%s' % (seq, stderr))

    os.remove(fname)
    return struct_list



import os, re, sys
import subprocess as sp
import random, string
import numpy as np
from .utils import *
from .pfunc import pfunc

DEBUG=False

# load package locations from yaml file, watch! global dict
package_locs = load_package_locations()

def free_energy(seq, constraint=None, package='vienna_2', T=37, coaxial=True, dna=False, beam_size=100,
		 pseudo=False, dangles=True, reweight=None, ensemble=True, param_file=None, linear=False,DEBUG=False):
	''' Compute free energy of RNA sequence. If structure is given, computes free energy of that structure. 
			Otherwise, returns MFE structure of sequence [NOT IMPLEMENTED YET].

		Args:
		seq (str): nucleic acid sequence
		constraint (str, optional): possible structure to constrain to in dot bracket notation
		T (float): temperature (Celsius), default 37

		ensemble (bool): to compute ensemble of constraint string or not.
			Just converts '.' to 'x' in string.
			If you want the free energy of just one structure,
			better practice is to use 'x' to denote unpaired. 


		motif (str): argument to vienna motif 
		beam_size (int): beam size for use in LinearPartition (Vienna, CONTRAfold, EternaFold only)
		dangles (bool): dangles or not, specifiable for vienna, nupack
                dna (bool): use SantaLucia model for DNA (NUPACK only)
		coaxial (bool): coaxial stacking or not, specifiable for rnastructure, vfold
		noncanonical(bool): include noncanonical pairs or not (for contrafold, RNAstructure (Cyclefold))
		pseudo (bool): include pseudoknot (nupack only)
		Implemented packages: 
		'vienna_1', 'vienna_2', 'contrafold'

		NB: doesn't multiply by kT for contrafold...
		
	Returns
		free energy (float)
	'''
	if not ensemble:
		constraint = constraint.replace('.','x')

	return pfunc(seq, package=package, T=T, dangles=dangles, coaxial=coaxial, pseudo=pseudo, dna=dna, beam_size = beam_size,
	 constraint=constraint, reweight=reweight, param_file=param_file, return_free_energy=True, linear=linear, DEBUG=DEBUG)

	# if package.lower().startswith('contrafold'):
	# 	Z_constrained = pfunc(seq, package=package, T=T, dangles=dangles, constraint=constraint,param_file=param_file)

	# 	return -1* np.log(Z_constrained) # .00198 is k in kcal/mol #0.0019899*(273+T) * 
	# else:
	# 	raise RuntimeError("%s `free_energy` not implemented yet" % package)



import numpy as np
import argparse, sys

def convert_dotbracket_to_matrix(s):
    m = np.zeros([len(s),len(s)])
    for char_set in [['(',')'], ['[',']'],['{','}'],['<','>']]:
        bp1=[]
        bp2=[]
        for i, char in enumerate(s):
            if char==char_set[0]:
                bp1.append(i)
            if char==char_set[1]:
                bp2.append(i)
        for i in list(reversed(bp1)):
            for j in bp2:
                if j > i:
                    m[i,j]=1.0
                    bp2.remove(j)
                    break
    return m



def convert_matrix_to_dotbracket(m):
    bp_list = convert_matrix_to_bp_list(m)
    return convert_bp_list_to_dotbracket(bp_list,len(m))

def convert_matrix_to_bp_list(m):
    bp_list = [] # convert adjacency matrix to adjacency list
    for i,row in enumerate(m):
        for j,is_bp in enumerate(row[i+1:]):
            if is_bp:
                bp_list.append([i,i+1+j])
    return bp_list


def convert_bp_list_to_dotbracket(bp_list,seq_len):
    dotbracket = "."*seq_len
    # group into bps that are not intertwined and can use same brackets!
    groups = group_into_non_conflicting_bp_(bp_list)

    # all bp that are not intertwined get (), but all others are
    # groups to be nonconflicting and then asigned (), [], {}, <> by group
    chars_set = [("(",")"),("(",")"),("[","]"),("{","}"),("<",">")]
    if len(groups) > len(chars_set):
        print("WARNING: PK too complex, not enough brackets to represent it.")

    for group,chars in zip(groups,chars_set):
        for bp in group:
            dotbracket = dotbracket[:bp[0]] + chars[0] + dotbracket[bp[0]+1:bp[1]] + chars[1] + dotbracket[bp[1]+1:]
    return dotbracket


def load_matrix_or_dbn(s):
    num_lines = sum(1 for line in open(s))

    if num_lines > 2: #heuristic here
        struct = np.loadtxt(s) # load as base pair matrix
        assert struct.shape[0] == struct.shape[1]
    else:
        try: # load as dot-bracket string

            dbn_struct = open(s,'r').read().rstrip()

            struct = convert_dotbracket_to_matrix(dbn_struct)
        except:
            raise ValueError('Unable to parse structure %s' % s)
    return struct

def score_ground_truth(pred_matrix, true_matrix):
    '''Score a predicted structure against a true structure,
     input as NxN base pair matrix (takes top triangle).'''

    N = pred_matrix.shape[0]
    #print('pred',pred_matrix.shape, 'true', true_matrix.shape)
    assert pred_matrix.shape[1] == N
    assert true_matrix.shape[0] == N
    assert true_matrix.shape[1] == N

    true = true_matrix[np.triu_indices(N)]
    pred = pred_matrix[np.triu_indices(N)]

    TP, FP, cFP, TN, FN = 0, 0, 0, 0, 0

    for i in range(len(true)):
        if true[i] == 1:
            if pred[i] == 1:
                TP += 1
            else:
                FN += 1
        elif true[i] == 0:
            if pred[i] == 0:
                TN += 1
            else:
                FP += 1
                #check for compatible false positive
                a,b = np.triu_indices(N)
                if np.sum(true_matrix,axis=0)[a[i]]+ np.sum(true_matrix,axis=0)[b[i]]==0:
                   cFP +=1

    # cFP = 0 #for debugging

    #print('TP', TP, 'TN', TN, 'FP', FP, 'FN', FN, 'cFP', cFP)

    if TP + FN == 0:
        sen = 1
    else:
        sen = TP/(TP + FN)

    if TP + FP - cFP == 0:
        ppv = 1
    else:
        ppv = TP/(TP + FP - cFP)

    mcc_num = (TP*TN - (FP - cFP)*FN)
    mcc_denom = np.sqrt((TP + FP - cFP)*(TP + FN)*(TN + FP - cFP)*(TN + FN))

    if  mcc_denom == 0:
        mcc = mcc_num
    else:
        mcc = mcc_num/mcc_denom

    if ppv + sen == 0:
        fscore = 0
    else:
        fscore = 2*ppv*sen/(ppv+sen)

    return sen, ppv, mcc, fscore, N


def group_into_non_conflicting_bp_(bp_list):
    ''' given a conflict list from get_list_bp_conflicts_, group basepairs into groups that do not conflict

    Args
        conflict_list: list of pairs of base_pairs that are intertwined basepairs

    Returns:
        groups of baspairs that are not intertwined
    '''
    conflict_list = get_list_bp_conflicts_(bp_list)

    non_redudant_bp_list = get_non_redudant_bp_list_(conflict_list)
    bp_with_no_conflict = [bp for bp in bp_list if bp not in non_redudant_bp_list]
    groups = [bp_with_no_conflict]
    while non_redudant_bp_list != []:
        current_bp = non_redudant_bp_list[0]
        current_bp_conflicts = []
        for conflict in conflict_list:
            if current_bp == conflict[0]:
                current_bp_conflicts.append(conflict[1])
            elif current_bp == conflict[1]:
                current_bp_conflicts.append(conflict[0])
        group = [bp for bp in non_redudant_bp_list if bp not in current_bp_conflicts]
        groups.append(group)
        non_redudant_bp_list = current_bp_conflicts
        conflict_list = [conflict for conflict in conflict_list if conflict[0] not in group and conflict[1] not in group]
    return groups


def get_list_bp_conflicts_(bp_list):
    '''given a bp_list gives the list of conflicts bp-s which indicate PK structure
    Args:
        bp_list: of list of base pairs where the base pairs are list of indeces of the bp in increasing order (bp[0]<bp[1])
    returns:
        List of conflicting basepairs, where conflicting is pairs of base pairs that are intertwined.
    '''
    if len(bp_list) <= 1:
        return []
    else:
        current_bp = bp_list[0]
        conflicts = []
        for bp in bp_list[1:]:
            if (bp[0] < current_bp[1] and current_bp[1] < bp[1]):
                conflicts.append([current_bp,bp])
        return conflicts + get_list_bp_conflicts_(bp_list[1:])

def get_non_redudant_bp_list_(conflict_list):
    ''' given a conflict list get the list of nonredundant basepairs this list has

    Args:
        conflict_list: list of pairs of base_pairs that are intertwined basepairs
    returns:
        list of basepairs in conflict list without repeats
    '''
    non_redudant_bp_list = []
    for conflict in conflict_list:
        if conflict[0] not in non_redudant_bp_list:
            non_redudant_bp_list.append(conflict[0])
        if conflict[1] not in non_redudant_bp_list:
            non_redudant_bp_list.append(conflict[1])
    return non_redudant_bp_list



from arnie.free_energy import free_energy
from arnie.mfe import mfe

seq = 'CGCUGUCUGUACUUGUAUCAGUACACUGACGAGUCCCUAAAGGACGAAACAGCG'
dG = free_energy(seq, linear=True, DEBUG=True)
print(dG)

dG = free_energy(seq, linear=True, package='contrafold', DEBUG=True)
print(dG)

dG = free_energy(seq, linear=True, package='eternafold', DEBUG=True)
print(dG)

struct = mfe(seq, linear=True)
print(struct)
struct = mfe(seq, linear=True, package='contrafold')
print(struct)
struct = mfe(seq, linear=True, package='eternafold')
print(struct)




import numpy as np
import argparse, sys
from arnie.mea.mea_utils import *
from copy import copy

class MEA:
    def __init__(self, bpps, gamma = 1.0, debug=False, run_probknot_heuristic = False, theta=0, stochastic=False):
        self.debug = debug
        self.bpps = bpps
        self.N=self.bpps.shape[0]
        self.gamma = gamma
        self.theta = theta
        self.W = np.zeros([self.N,self.N])
        self.MEA_bp_list = []
        self.structure = ['.']*self.N
        self.MEA_bp_matrix = np.zeros([self.N, self.N])
        self.tb = np.zeros([self.N, self.N])
        self.min_hp_length = 3
        self.evaluated = False
        self.stochastic = stochastic

        if run_probknot_heuristic:
            self.run_ProbKnot()
        else:
            self.run_MEA()
        
    def fill_W(self, i, j):
        if self.stochastic:
            options = [self.W[i+1, j], self.W[i, j-1],\
             (self.gamma+1)*self.bpps[i,j] + self.W[i+1, j-1] - 1,\
            np.max([self.W[i,k] + self.W[k+1, j] for k in range(i+1,j)])]
            option_wts = options - np.min(options)
            option_wts /= np.sum(option_wts)
            selection = np.random.choice([0,1,2,3],p=option_wts)
            self.W[i,j] = options[selection]
            self.tb[i,j] = selection #0: 5' pass, 1: 3' pass, 2: bp, 3: multiloop

        else:
            options = [self.W[i+1, j], self.W[i, j-1],\
             (self.gamma+1)*self.bpps[i,j] + self.W[i+1, j-1] - 1,\
            np.max([self.W[i,k] + self.W[k+1, j] for k in range(i+1,j)])]
            self.W[i,j] = np.max(options) 
            self.tb[i,j] = np.argmax(options) #0: 5' pass, 1: 3' pass, 2: bp, 3: multiloop
            
    def run_MEA(self):
        # fill weight matrix
        for length in range(self.min_hp_length, self.N):
            for i in range(self.N-length):
                j = i + length
                self.fill_W(i,j)
                
        self.traceback(0,self.N-1)
        
        for x in self.MEA_bp_list:
            self.MEA_bp_matrix[x[0],x[1]]=1
            self.structure[x[0]]='('
            self.structure[x[1]]=')'
        
        self.structure = ''.join(self.structure)
        if not self.evaluated: self.evaluated = True

    def run_ProbKnot(self):

        #Threshknot step: filter out bps below cutoff theta
        threshknot_filter = np.where(self.bpps <= self.theta)
        filtered_bpps = copy(self.bpps)
        filtered_bpps[threshknot_filter] = 0

        output = np.zeros([self.N, self.N])
        
        # ProbKnot heuristic part 1: get all base pairs where p(ij) == p_max(i)
        output[np.where(self.bpps == np.max(self.bpps, axis=0))] = 1
        
        # ProbKnot heuristic part 2: get all base pairs where p(ij) == p_max(j)
        self.MEA_bp_matrix = np.clip(output+np.transpose(output)-1,0,1)

        for [i, j] in np.array(np.where(self.MEA_bp_matrix == 1)).T:
            if np.abs(i - j) > 1:
                if [j,i] not in self.MEA_bp_list:
                    self.MEA_bp_list.append([i,j])
                    #self.structure[i] = '('
                    #self.structure[j] = ')'
        #print('Warning: formatting pseudoknotted dot-bracket structures not yet supported. Any pseudoknotted stems will only appear as parentheses (not brackets).')
        #self.structure = ''.join(self.structure)
        self.structure = convert_bp_list_to_dotbracket(self.MEA_bp_list,len(self.bpps))

        if not self.evaluated: self.evaluated = True

    def traceback(self, i, j):
        if j <= i:
            return
        elif self.tb[i,j] == 0: #5' neighbor
            if self.debug: print(i,j, "5'")
            self.traceback(i+1,j)
        elif self.tb[i,j] == 1: #3' neighbor
            if self.debug: print(i,j, "3'")
            self.traceback(i,j-1)
        elif self.tb[i,j] == 2: # base pair
            if self.debug: print(i,j,'bp')
            self.MEA_bp_list.append((i,j))
            self.traceback(i+1,j-1)
        else: #multiloop
            for k in range(i+1,j):
                if self.W[i,j] == self.W[i, k] + self.W[k+1,j]:
                    if self.debug: print(i,j,"multiloop, k=",k)
                    self.traceback(i,k)
                    self.traceback(k+1,j)
                    break

    def score_expected(self):
        '''Compute expected values of TP, FP, etc from predicted MEA structure.

         Returns: 
         pseudoexpected SEN, PPV, MCC, F-score'''

        if not self.evaluated: 
            if run_probknot_heuristic:
                self.run_ProbKnot()
            else:
                self.run_MEA()

        pred_m = self.MEA_bp_matrix[np.triu_indices(self.N)]
        probs = self.bpps[np.triu_indices(self.N)]

        TP = np.sum(np.multiply(pred_m, probs)) + 1e-6
        TN = 0.5*self.N*self.N-1 - np.sum(pred_m) - np.sum(probs) + TP + 1e-6
        FP = np.sum(np.multiply(pred_m, 1-probs)) + 1e-6
        FN = np.sum(np.multiply(1-pred_m, probs)) + 1e-6

        a,b = np.triu_indices(self.N)
        cFP = 1e-6
        # for i in range(len(pred_m)):
        #     if np.sum(self.MEA_bp_matrix,axis=0)[a[i]] + np.sum(self.MEA_bp_matrix,axis=0)[b[i]]==0:
        #        cFP += np.multiply(pred_m[i], 1-probs[i])

        sen = TP/(TP + FN)
        ppv = TP/(TP + FP - cFP)
        mcc = (TP*TN - (FP - cFP)*FN)/np.sqrt((TP + FP - cFP)*(TP + FN)*(TN + FP - cFP)*(TN + FN))
        fscore = 2*TP/(2*TP + FP - cFP + FN)

        return [sen, ppv, mcc, fscore]

    def score_ground_truth(self, ground_truth_struct, allow_pseudoknots=False):
        if len(ground_truth_struct[0])==1:
            gt_matrix = convert_dotbracket_to_matrix(ground_truth_struct)
        else:
            gt_matrix = ground_truth_struct

        if not self.evaluated: self.run_MEA()
        sen, ppv, mcc, fscore, _ = score_ground_truth(self.MEA_bp_matrix, gt_matrix)
        return [sen, ppv, mcc, fscore]



import os, re, sys, shutil
import subprocess as sp
import random, string
import numpy as np
from .utils import *

DEBUG=False

# load package locations from yaml file, watch! global dict
package_locs = load_package_locations()

def mfe(seq, package='vienna_2', T=37,
    constraint=None, motif=None,
    linear=False, return_dG_MFE = False,
    dangles=True, noncanonical=False, beam_size=100,
    bpps=False, param_file=None, coaxial=True, reweight=None,viterbi = False,
    probing_signal=None,probing_kws=None, pseudo=False,
    shape_signal=None, dms_signal=None, shape_file=None, dms_file=None, **kwargs):

    ''' Compute MFE structure (within package) for RNA sequence.
    Note: this is distinct from the arnie MEA codebase, which takes any base pair probability matrix and computes the maximum expected accuracy structure.
    That said, Contrafold's default structure prediction is an MEA structure, not MFE.  In this module, calling Contrafold returns the default MEA structure unless the 
    --viterbi flag is used, which will do the viterbi (MFE) algorithm in contrafold.

        Args:
        seq (str): nucleic acid sequence
        T (float): temperature (Celsius)
        constraint (str): structure constraints
        linear (bool): call LinearFold to estimate MFE in Vienna or Contrafold
        motif (str): argument to vienna motif 
        return_dG_MFE (bool): also return dG(MFE) (specific to linearfold)
        dangles (bool): dangles or not, specifiable for vienna, nupack
        coaxial (bool): coaxial stacking or not, specifiable for rnastructure, vfold
        noncanonical(bool): include noncanonical pairs or not (for contrafold, RNAstructure (Cyclefold))
        shape_signal(list): list of normalized SHAPE reactivities, with negative values indicating no signal
        dms_signal(list): list of normalized DMS reactivities, with negative values indicating no signal
        pseudo: if True, will predict pseudoknots

        Possible packages: 
        'vienna_2', 'vienna_1','contrafold_1','contrafold_2', 'rnastructure'
        
    Returns
        string: MFE structure
    '''

    # TODO: update to be just probing_signal
    # if shape_signal:
    #     print('Warning: shape_signal is deprecated, use probing_signal')
    #     probing_signal = copy(shape_signal)

    # if dms_signal:
    #     print('Warning: dms_signal is deprecated, use probing_signal')
    #     probing_signal = copy(dms_signal)

    try:
        pkg, version = package.lower().split('_')
    except:
        pkg, version = package.lower(), None

    if not bpps: # if bpps, already printed these warnings
        if not dangles and pkg not in ['vienna', 'nupack']:
            print('Warning: %s does not support dangles options' % pkg)
        if not coaxial and pkg not in ['rnastructure', 'vfold']:
            print('Warning: %s does not support coaxial options' % pkg)

    if linear and pkg not in ['vienna','contrafold','eternafold']:
        print('Warning: LinearFold only implemented for vienna, contrafold, eternafold.')

    if pseudo and not pkg in ['rnastructure', 'nupack']:
        print('Warning: %s and pseudoknots not supported in Arnie yet' % pkg)

    if pkg=='vienna':
        if linear:
            if return_dG_MFE:
                struct, dG_MFE = mfe_linearfold_(seq, package='vienna', return_dG_MFE=return_dG_MFE, beam_size=beam_size)
            else:
                struct = mfe_linearfold_(seq, package='vienna', return_dG_MFE=return_dG_MFE)
        else:
            struct = mfe_vienna_(seq, version=version, T=T, dangles=dangles, constraint=constraint, motif=motif, param_file=param_file,
                reweight=reweight, probing_signal=probing_signal, **kwargs)
 
    elif pkg=='contrafold':
        if linear:
            if return_dG_MFE:
                struct, dG_MFE = mfe_linearfold_(seq, package='contrafold', return_dG_MFE=return_dG_MFE, beam_size=beam_size)
            else:
                struct = mfe_linearfold_(seq, package='contrafold', return_dG_MFE=return_dG_MFE)
        else:
            struct = mfe_contrafold_(seq, version=version, T=T, constraint=constraint, param_file=param_file,viterbi=viterbi)

    elif pkg=='eternafold':
        if linear:
            if return_dG_MFE:
                struct, dG_MFE = mfe_linearfold_(seq, package='eternafold', return_dG_MFE=return_dG_MFE, beam_size=beam_size)
            else:
                struct = mfe_linearfold_(seq, package='eternafold', return_dG_MFE=return_dG_MFE)

        else:

            if 'eternafoldparams' in package_locs.keys() and 'eternafold' not in package_locs.keys():
                struct = mfe_contrafold_(seq, version=version, T=T, constraint=constraint, param_file=package_locs['eternafoldparams'],viterbi=viterbi)

            elif 'eternafold' in package_locs.keys():
                #Using eternafold code and params in eternafold codebase
                efold_param_file = package_locs['eternafold']+'/../parameters/EternaFoldParams.v1'
                print(efold_param_file)
                print(os.path.exists(efold_param_file))
                if not os.path.exists(efold_param_file):
                    raise RuntimeError('Error: Parameters not found at %s' % efold_param_file)
                else:
                    struct = mfe_contrafold_(seq, version=version, T=T, constraint=constraint, DIRLOC=package_locs['eternafold'],
                        param_file=efold_param_file,viterbi=viterbi, probing_signal=probing_signal, probing_kws=probing_kws)


    elif pkg=='rnastructure':
        if linear:
            raise ValueError('package %s is not supported with linearfold.' % package)
        else:
            struct = mfe_rnastructure_(seq, version=version, T=T, constraint=constraint, 
                probing_signal=probing_signal,
                param_file=param_file, shape_signal=shape_signal, dms_signal=dms_signal, 
                shape_file=shape_file, dms_file=dms_file, pseudo = pseudo)
    else:
        raise ValueError('package %s not understood.' % package)

    if return_dG_MFE:
        return struct, dG_MFE
    else:
        return struct

def mfe_vienna_(seq, T=37, version='2', constraint=None, motif=None, param_file=None, dangles=True, reweight=None,
    probing_signal=None, shapeMethod='W', probing_kws=None, **kwargs):
    """get minimum free energy structure with Vienna

    Args:
        seq (str): nucleic acid sequence
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif  
    Returns
        str: secondary structure representation for MFE
    """

    if not version:
        version='2'

    if version.startswith('2'):
        LOC=package_locs['vienna_2']
    elif version.startswith('1'):

        LOC=package_locs['vienna_1']

    else:
        raise RuntimeError('Error, vienna version %s not present' % version)

    if constraint is not None:
        fname = write([seq, constraint])
        command.append('-C')
        command.append('--enforceConstraint')
    else:
        fname = write([seq])

    command = ['%s/RNAfold' % LOC, '-T', str(T), '-p0'] #p0 doesn't predict bpps, saves time
    if motif is not None:
        command.append('--motif=%s' % motif)

    if probing_signal is not None:
        if probing_kws is None:
            probing_kws={}

        if shapeMethod=='W':
            probing_file = run_RNAPVmin(probing_signal, seq, LOC, DEBUG, **probing_kws)

        elif shapeMethod=='D' or shapeMethod=='Z':
            probing_file = write_reactivity_file_RNAstructure(probing_signal)
            command.append('--shapeConversion=O')

        command.append('--shape=%s' % probing_file)
        command.append('--shapeMethod=%s' % shapeMethod)

    if not dangles:
        command.append('--dangles=0')
        
    if reweight is not None:
        command.append('--commands=%s' % reweight)

    if param_file:
        command.append('--paramFile=%s' % param_file)

    with open(fname) as f:
        if DEBUG: print(fname)
        if DEBUG: print(' '.join(command))
        p = sp.Popen(command, stdin=f, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if p.returncode:
        raise Exception('RNAfold failed: on %s\n%s' % (seq, stderr))
    os.remove(fname)
    os.remove('rna.ps')

    if 'omitting constraint' in stderr.decode('utf-8'):
        raise ValueError('Constraint caused impossible structure')
    else:
        return stdout.decode('utf-8').split('\n')[1].split(' ')[0]

def mfe_rnastructure_(seq, T=24, version=None, constraint=None, param_file=None, probing_signal=None,probing_kws=None,
    shape_signal=None, dms_signal=None, shape_file=None, dms_file=None, pseudo=False):
    """get minimum free energy structure
        with SHAPE or DMS data, uses the default slope and intercept in RNAStructure

    Args:
        seq (str): nucleic acid sequence
        T (float): temperature
    Returns
        float: MFE structure
    """

    if probing_signal is not None:
        shape_signal = probing_signal

    if param_file is not None:
        raise ValueError('Cannot run RNAstructure with non-default RNA parameters as specified in: %s' % param_file)
    if version is not None:
        raise ValueError('Cannot run RNAstructure with non-default version: %s' % version)
    if (shape_signal is not None) and (shape_file is not None):
        raise ValueError('Please specify SHAPE reactivities either as a list or in a SHAPE reactivity file')
    if (dms_signal is not None) and (dms_file is not None):
        raise ValueError('Please specify DMS reactivities either as a list or in a DMS reactivity file')

    LOC=package_locs['rnastructure']

    seq_file = write(['>sequence', seq])
    ct_fname = '%s.ct' % filename()

    command = []
    if not pseudo:
        command = command + ['%s/Fold' % LOC, seq_file, ct_fname, '-T', str(T + 273.15)]
    else:
        command = command + ['%s/ShapeKnots' % LOC, seq_file, ct_fname]
        # if dms_signal is not None:
        #     raise ValueError('Cannot run RNAstructure with DMS signal and pseudoknots.')
        if constraint is not None:
            raise ValueError('Cannot run RNAstructure with constraints and pseudoknots.')
    
    con_fname = None
    dms_fname = None
    shape_fname = None

    if constraint is not None:
        con_fname = '%s.CON' % filename()
        convert_dbn_to_RNAstructure_input(seq, constraint, con_fname)
        command.extend(['--constraint', con_fname])

    if dms_signal is not None:
        if len(dms_signal) != len(seq):
            raise RuntimeError('DMS signal used with RNAstructure must have same length as the sequence.')
        dms_fname = write_reactivity_file_RNAstructure(dms_signal)
        command.extend(['--DMS', dms_fname])

    if dms_file is not None:
        command.extend(['--DMS', dms_file])

    if shape_signal is not None:
        if len(shape_signal) != len(seq):
            raise RuntimeError('SHAPE signal used with RNAstructure must have same length as the sequence.')
        shape_fname = write_reactivity_file_RNAstructure(shape_signal)
        command.extend(['--SHAPE', shape_fname])

    if shape_file is not None:
        command.extend(['--SHAPE', shape_file])

    if DEBUG: print(' '.join(command))

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('RNAstructure failed: on %s\n%s' % (seq, stderr))

    if con_fname is not None:
        os.remove(con_fname)
    if dms_fname is not None:
        os.remove(dms_fname)
    if shape_fname is not None:
        os.remove(shape_fname)
    if seq_file is not None:
        os.remove(seq_file)

    dot_fname = '%s.dbn' % filename()
    command = ['%s/ct2dot' % LOC, ct_fname, "1", dot_fname]

    if DEBUG: print(' '.join(command))

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('RNAstructure ct2dot failed: on %s\n%s' % (seq, stderr))

    f = open(dot_fname)
    dot_lines = f.readlines()        
    f.close()

    mfe_struct = dot_lines[-1].strip('\n')

    os.remove(ct_fname)
    os.remove(dot_fname)

    return mfe_struct

def mfe_contrafold_(seq, T=37, version='2', constraint=None, param_file=None,DIRLOC=None,
    viterbi=False, probing_signal=None, probing_kws=None):
    """get MFE structure for Contrafold

    Args:
        seq (str): nucleic acid sequence
        T (float): temperature
        constraint (str): structure constraints
        motif (str): argument to vienna motif  
    Returns
        secondary structure dot-bracket string for MFE
    """
    if not version: version='2'

    if probing_signal is not None:
        fname = write_reactivity_file_contrafold(probing_signal, seq)
    else:
        fname = '%s.in' % filename()

    if DIRLOC is not None:
        LOC=DIRLOC
    elif version.startswith('2'):
        LOC=package_locs['contrafold_2']
    elif version.startswith('1'):
        LOC=package_locs['contrafold_1']
    else:
        raise RuntimeError('Error, Contrafold version %s not present' % version)

    command = ['%s/contrafold' % LOC, 'predict', fname]

    if probing_signal is not None:
        command = command + ['--evidence', '--params', package_locs['eternafold']+'/../parameters/EternaFoldParams_PLUS_POTENTIALS.v1', '--numdatasources','1', ]
        if probing_kws is not None:
            if 'kappa' in probing_kws.keys():
                command = command + ['--kappa', str(probing_kws['kappa']) ]
    else:
        if param_file is not None:
            command = command + ['--params', param_file]

    if viterbi:
        command.append('--viterbi')

    if constraint is not None:
        convert_dbn_to_contrafold_input(seq, constraint, fname)
        command.append('--constraints')
    else:
        if probing_signal is None:
            convert_dbn_to_contrafold_input(seq, ''.join(['.' for x in range(len(seq))]), fname)

    if DEBUG: print(' '.join(command))

    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)
    if p.returncode:
        raise Exception('Contrafold failed: on %s\n%s' % (seq, stderr))

    os.remove(fname)
    
    return stdout.decode('utf-8').split('\n')[-2]

def mfe_linearfold_(seq, bpps=False, package='contrafold', beam_size=100, return_dG_MFE=False):
    
    seqfile = write([seq])

    LOC = package_locs['linearfold']

    if bpps:

        pf_only = 0
    else:
        pf_only = 1

    # args:  beamsize, is_sharpturn, is_verbose, is_eval, is_constraints]
    #Todo: implement constraint input
    command=['echo %s | %s/linearfold_%s' % (seq, LOC, package[0]), str(beam_size), '0', '0', '0']
    if DEBUG: print(' '.join(command))
    p = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)

    stdout, stderr = p.communicate()

    if DEBUG:
        print('stdout')
        print(stdout)
        print('stderr')
        print(stderr)

    if p.returncode:
        raise Exception('LinearFold failed: on %s\n%s' % (seq, stderr))


    # linearfold returns two different things depending on which package
    struct = stdout.decode('utf-8').split('\n')[1].split(' ')[0]

    os.remove(seqfile)

    if return_dG_MFE:

        dG_mfe = float(stdout.decode('utf-8').split('\n')[1].split(' ')[1][1:-1])

        if package.lower() != 'vienna':
            dG_mfe *= -1

        return struct, dG_mfe

    else:
        return struct





import os
import re
import subprocess as sp
import random
import string
import numpy as np
import arnie


def complement_to_(string):
    base_pairing_dct = {'a': 'u', 'u': 'a', 'g': 'c', 'c': 'g', 't': 'a'}
    return ''.join(base_pairing_dct[x.lower()] for x in string[::-1])


def run_RNAPVmin(probing_signal, seq, LOC, DEBUG, tauSigmaRatio=1, shapeConversion='S'):
    reac_file = write_reactivity_file_vienna(probing_signal, seq)
    fname = write([seq])
    RNApvmin_command = ['%s/RNApvmin' % LOC, reac_file, '--shapeConversion=%s' %
                        shapeConversion, '--tauSigmaRatio=%f' % tauSigmaRatio]

    with open(fname) as f:
        if DEBUG:
            print(fname)
        if DEBUG:
            print(' '.join(RNApvmin_command))
        p = sp.Popen(RNApvmin_command, stdin=f, stdout=sp.PIPE, stderr=sp.PIPE)
    rnapvmin_stdout, rnapvmin_stderr = p.communicate()

    shape_file = filename()

    with open(shape_file, 'wb') as f:
        f.write(rnapvmin_stdout)

    if DEBUG:
        print('stdout')
        print(rnapvmin_stdout)
        print('stderr')
        print(rnapvmin_stderr)

    if p.returncode:
        raise Exception('RNApvmin failed: on %s\n%s' % (seq, stderr))

    return shape_file


###############################################################################
# File writing
###############################################################################


def convert_dbn_to_RNAstructure_input(seq, constraints, filename):
    assert(len(seq) == len(constraints))

    bp_list = convert_dotbracket_to_bp_dict(constraints)

    SS_list, pairs_list = [], []

    for i, (s, c) in enumerate(list(zip(seq, constraints))):
        if c == 'x':
            SS_list.append(i+1)
        elif c == '.':
            pass
        elif c == '(':
            pairs_list.append([i+1, bp_list[i]+1])
        elif c == ')':
            pass
        else:
            print('Error reading constraint string', c)

    with open('%s' % filename, 'w') as out:
        out.write('DS:\n-1\n')
        out.write('SS:\n')
        for x in SS_list:
            out.write('%d\n' % x)
        out.write('-1\n')
        out.write('Mod:\n-1\n')
        out.write('Pairs:\n')
        for x, y in pairs_list:
            out.write('%d %d\n' % (x, y))
        out.write('-1 -1')
        out.write('FMN:\n-1\nForbids:\n-1\n')


def write_vector_to_file(vector, outfile):
    for x in vector:
        outfile.write('%.3f\n' % x)
    return


def write_matrix_to_file(matrix, outfile):
    for x in matrix:
        outfile.write('\t'.join(['%.3f' % y for y in x])+'\n')
    return

def write_reactivity_file_RNAstructure(reactivities, fname=None):
    """ writes reactivities (either SHAPE or DMS) to file format used by RNAstructure

      ex:
      1 0.120768
      2 0.190510
      3 0.155776

    Args:
      reactivities (list): a list of normalized reactivity float data. 
      Negative numbers can be used to indicate no signal.
    """

    if fname is None:
        fname = '%s.SHAPE' % filename()
    with open(fname, 'w') as f:
        i = 1
        for reactivity in reactivities:
            if reactivity >= 0:
                f.write('%d %f\n' % (i, reactivity))
            i += 1
    return fname


def write_reactivity_file_contrafold(reactivities, sequence, fname=None):
    '''write reactivity (either SHAPE or DMS) to file format used by CONTRAfold.

      ex:
      1 U e1 0.120768
      2 G e1 0.190510
      3 U e1 0.155776

    Args:
      reactivities (list): a list of normalized reactivity float data. 
      sequence: RNA sequence
      Negative numbers can be used to indicate no signal.
    '''

    assert len(reactivities) == len(sequence)

    if fname is None:
        fname = '%s.bpseq' % filename()

    with open(fname, 'w') as f:
        i = 1
        for char, reactivity in list(zip(sequence, reactivities)):
            if reactivity > 0:
                f.write('%d %s e1 %.6f\n' % (i, char, reactivity))
            elif reactivity == 0:
                f.write('%d %s e1 0.0001\n' % (i, char))
            else:
                f.write('%d %s e1 0.0\n' % (i, char))

            i += 1
    return fname


def local_rand_filename(n=6):
    """generate random filename

    Args:
      n (int): number of characters
    """
    rand = ''.join([random.choice(string.ascii_lowercase) for _ in range(n)])
    return rand


def get_random_folder(n=6):
    """ generate randome foldername
    that does not exist in current folder"""
    out_folder = ''.join([random.choice(string.ascii_lowercase) for _ in range(n)])
    while os.path.isdir(out_folder):
        out_folder = ''.join([random.choice(string.ascii_lowercase) for _ in range(n)])
    tmpdir = load_package_locations()['TMP']
    out_folder = f'{tmpdir}/{out_folder}'
    return out_folder


def filename(n=6):
    """generate random filename

    Args:
      n (int): number of characters
    """
    rand = ''.join([random.choice(string.ascii_lowercase) for _ in range(n)])
    tmpdir = load_package_locations()['TMP']
    return '%s/%s' % (tmpdir, rand)


def write(lines, fname=None):
    """write lines to file

    Args:
      lines (list): line(s) to write to file
      fname (str): filename to write to
    """
    if fname is None:
        fname = '%s.in' % filename()
    with open(fname, 'w') as f:
        for line in lines:
            f.write('%s\n' % line)
    return fname

def convert_dbn_to_contrafold_input(seq, constraints, filename):
    constraint_list = write_constraint_string(seq, constraints)
    with open('%s' % filename, 'w') as out:
        for i in range(len(seq)):
            out.write('%d\t%s\t%d\n' % (i+1, seq[i], constraint_list[i]))


###############################################################################
# File reading
###############################################################################

def bpseq_to_bp_list(bpseq_file, header_length=1):
    ''' read a bpseq file into a bp_list
    assumes first line header_length is title 
    and starts reading only after that
    '''
    bps = open(bpseq_file).readlines()
    bps = bps[header_length:]
    bp_list = []
    for bp in bps:
        bp = bp.split()[::2]
        bp = [int(nt) for nt in bp]
        if bp[1] != 0 and bp[0] < bp[1]:
            bp_list.append([bp[0] - 1, bp[1] - 1])
    return bp_list


def ct_to_bp_list(ct_file, header_length=1):
    ''' read a ct file into a bp_list
    assumes first line header_length is title 
    and starts reading only after that
    '''
    bps = open(ct_file).readlines()
    bps = bps[header_length:]
    bp_list = []
    for bp in bps:
        bp = bp.split()[::4]
        if len(bp) != 0:
            bp = [int(nt) for nt in bp]
            if bp[1] != 0 and bp[0] < bp[1]:
                bp_list.append([bp[0] - 1, bp[1] - 1])
    return bp_list


def prob_to_bpp(prob_file):
    ''' read a .prob file and return a bpp
    '''
    return np.loadtxt(prob_file)


###############################################################################
# Package handling
###############################################################################

supported_packages = [
    "contrafold",
    "eternafold",
    "nupack",
    "RNAstructure",
    "RNAsoft",
    "vienna"
    # PK Predictors
    "e2efold",
    "hotknots", 
    "ipknot", 
    "knotty", 
    "pknots",
    "spotrna",
    "spotrna2",
]

def print_path_files():
    package_dct = load_package_locations()
    for key, v in package_dct.items():
        print(key, v)


def package_list():
    pkg_list = []
    package_dct = load_package_locations()
    for key, v in package_dct.items():
        if key != "TMP" and key.lower() != 'bprna':
            if not key.startswith('linear'):
                if key == 'eternafoldparams' and 'eternafold' not in pkg_list:
                    pkg_list.append('eternafold')
                else:
                    if v != "None":
                        pkg_list.append(key)
    return pkg_list


def load_package_locations(DEBUG=False):
    '''Set up  paths to RNA folding packages. Checks environment variables or a user-supplied file. If using, specify this in your ~/.bashrc as $ARNIEFILE'''
    return_dct = {}
    package_path = os.path.dirname(arnie.__file__)

    if DEBUG:
        print('Reading Arnie file at %s' % os.environ['ARNIEFILE'])
        print(supported_packages)

    # Read from environment variables
    for package in supported_packages:
        envVar = f"{package.upper()}_PATH"
        # Nupack installation sets its own environment variables
        if package == "nupack":
            envVar = f"{package.upper()}HOME"
        path = os.environ.get(envVar)
        if path:
            return_dct[package] = path

    # Read from Arnie file as last resort
    with open("%s" % os.environ["ARNIEFILE"], 'r') as f:
        for line in f.readlines():
            if line.strip():
                if not line.startswith('#'):
                    key, string = line.split(':')
                    string = string.strip()
                    if key not in return_dct:
                        return_dct[key] = string

    return return_dct


###############################################################################
# Structure representation conversion
###############################################################################

def convert_dotbracket_to_bp_list(s, allow_pseudoknots=False):
    if allow_pseudoknots:
        lefts = ["(", "[", "{", "<"]
        rights = [")", "]", "}", ">"]
        lower_alphabet = [chr(lower) for lower in range(97, 123)]
        upper_alphabet = [chr(upper) for upper in range(65, 91)]
        lefts.extend(lower_alphabet)
        rights.extend(upper_alphabet)
    else:
        lefts = ["("]
        rights = [")"]

    char_ignored = [char for char in s if char not in rights + lefts + ["."]]
    char_ignored = list(set(char_ignored))
    if char_ignored != []:
        print("WARNING: characters in structure,", char_ignored, "ignored!")

    l = []
    for left, right in zip(lefts, rights):
        bp1 = []
        bp2 = []
        for i, char in enumerate(s):
            if char == left:
                bp1.append(i)
            elif char == right:
                bp2.append(i)

        for i in list(reversed(bp1)):
            for j in bp2:
                if j > i:
                    l.append([i, j])

                    bp2.remove(j)
                    break
    l = sorted(l, key=lambda x: x[0])
    return l


def convert_dotbracket_to_bp_dict(s, allow_pseudoknots=False):

    if allow_pseudoknots:
        lefts = ["(", "[", "{", "<"]
        rights = [")", "]", "}", ">"]
        lower_alphabet = [chr(lower) for lower in range(97, 123)]
        upper_alphabet = [chr(upper) for upper in range(65, 91)]
        lefts.extend(lower_alphabet)
        rights.extend(upper_alphabet)
    else:
        lefts = ["("]
        rights = [")"]

    char_ignored = [char for char in s if char not in rights + lefts + ["."]]
    char_ignored = list(set(char_ignored))
    if char_ignored != []:
        print("WARNING: characters in structuture,", char_ignored, "ignored!")

    m = {}
    for left, right in zip(lefts, rights):
        bp1 = []
        bp2 = []
        for i, char in enumerate(s):
            if char == left:
                bp1.append(i)
            elif char == right:
                bp2.append(i)

        for i in list(reversed(bp1)):
            for j in bp2:
                if j > i:
                    m[i] = j
                    m[j] = i

                    bp2.remove(j)
                    break

    return m


def convert_dotbracket_to_matrix(s, allow_pseudoknots=False):
    matrix = np.zeros([len(s), len(s)])
    bp_list = convert_dotbracket_to_bp_dict(
        s, allow_pseudoknots=allow_pseudoknots)
    for k, v in bp_list.items():
        matrix[k, v] = 1
    return matrix


def convert_bp_list_to_dotbracket(bp_list, seq_len):
    db = "." * seq_len
    # group into bps that are not intertwined and can use same brackets!
    groups = _group_into_non_conflicting_bp(bp_list)

    # all bp that are not intertwined get (), but all others are
    # groups to be nonconflicting and then asigned (), [], {}, <> by group
    chars_set = [("(", ")"), ("(", ")"), ("[", "]"), ("{", "}"), ("<", ">")]
    alphabet = [(chr(lower), chr(upper))
                for upper, lower in zip(list(range(65, 91)), list(range(97, 123)))]
    chars_set.extend(alphabet)

    if len(groups) > len(chars_set):
        print("WARNING: PK too complex, not enough brackets to represent it.")

    for group, chars in zip(groups, chars_set):
        for bp in group:
            db = db[:bp[0]] + chars[0] + \
                db[bp[0] + 1:bp[1]] + chars[1] + db[bp[1] + 1:]
    return db


def get_bpp_from_dbn(dbn_struct):
    """ Gets a base-pairing matrix from a dot-bracket notation structure
    The matrix has 1's at position i,j if sequence positions i,j are base-paired
    and 0 otherwise.

    Args: 
      dbn_struct - structure in MFE format
    Return:
      2D numpy array
    """

    bpp_matrix = np.zeros((len(dbn_struct), len(dbn_struct)))

    bkt_open = "(<{"
    bkt_close = ")>}"

    for ii in range(len(bkt_open)):
        open_char = bkt_open[ii]
        close_char = bkt_close[ii]

        open_pos = []

        for jj in range(len(dbn_struct)):
            if dbn_struct[jj] == open_char:
                open_pos += [jj]
            if dbn_struct[jj] == close_char:
                pair = open_pos[-1]
                del open_pos[-1]
                bpp_matrix[jj, pair] = 1
                bpp_matrix[pair, jj] = 1

    return bpp_matrix

def get_helices(s, allowed_buldge_len=0):
    bp_list = convert_dotbracket_to_bp_list(s, allow_pseudoknots=True)
    helices = []
    current_helix = []
    while bp_list != []:
        current_bp = bp_list.pop(0)
        if current_helix == []:
            current_helix.append(current_bp)
        else:
            in_helix_left = list(range(current_helix[-1][0] + 1, current_helix[-1][0] + allowed_buldge_len + 2))
            in_helix_right = list(range(current_helix[-1][1] - allowed_buldge_len - 1, current_helix[-1][1]))
            if current_bp[0] in in_helix_left and current_bp[1] in in_helix_right:
                current_helix.append(current_bp)
            else:
                helices.append(current_helix)
                current_helix = [current_bp]
    helices.append(current_helix)
    return helices

def post_process_struct(s, allowed_buldge_len=0, min_len_helix=1):
    ''' given a structure, remove any helices that are too short

    allowed_buldge_len, if 0 does not allow any buldges when calculating the length
    of the helices, if 1 allows 0-1 and 1-1 buldges, if 2 additionally allows 2-0,2-1,2-2 bulges etc
    min_len_helix, any helices less than this value will be removed.
    '''
    helices = get_helices(s, allowed_buldge_len)
    bp_list_out = []
    for helix in helices:
        if len(helix) >= min_len_helix:
            bp_list_out.extend(helix)
    s_out = convert_bp_list_to_dotbracket(bp_list_out, len(s))
    return s_out


###############################################################################
# Evaluating a structure
###############################################################################

def get_expected_accuracy(dbn_string, bp_matrix, mode='mcc'):
    '''given a secondary structure as dbn string and base pair matrix, 
    assess expected accuracy for the structure.

    Inputs:
    dbn_string (str): Secondary structure string in dot-parens notation.
    bp_matrix (NxN array):  symmetric matrix of base pairing probabilities.
    mode: ['mcc','fscore','sen','ppv']: accuracy metric for which to compute expected value.

    Returns: expected accuracy value.
    '''

    assert bp_matrix.shape[0] == bp_matrix.shape[1]
    assert bp_matrix.shape[0] == len(dbn_string)

    struct_matrix = convert_dotbracket_to_matrix(dbn_string)
    N = len(dbn_string)

    pred_m = struct_matrix[np.triu_indices(N)]
    probs = bp_matrix[np.triu_indices(N)]

    TP = np.sum(np.multiply(pred_m, probs)) + 1e-6
    TN = 0.5*N*N-1 - np.sum(pred_m) - np.sum(probs) + TP + 1e-6
    FP = np.sum(np.multiply(pred_m, 1-probs)) + 1e-6
    FN = np.sum(np.multiply(1-pred_m, probs)) + 1e-6

    a, b = np.triu_indices(N)
    cFP = 1e-6  # compatible false positives
    # for i in range(len(pred_m)):
    #     if np.sum(struct_matrix,axis=0)[a[i]] + np.sum(struct_matrix,axis=0)[b[i]]==0:
    #        cFP += np.multiply(pred_m[i], 1-probs[i])

    if mode == 'sen':
        return TP/(TP + FN)
    elif mode == 'ppv':
        return TP/(TP + FP - cFP)
    elif mode == 'mcc':
        return (TP*TN - (FP - cFP)*FN)/np.sqrt((TP + FP - cFP)*(TP + FN)*(TN + FP - cFP)*(TN + FN))
    elif mode == 'fscore':
        return 2*TP/(2*TP + FP - cFP + FN)
    else:
        print('Error: mode not understood.')


def get_mean_base_pair_propensity(dbn_string):
    '''Measure of base pair locality.'''
    mat = convert_dotbracket_to_matrix(dbn_string)
    i, j = np.where(mat == 1)
    # divide by 2 because symmetric matrix
    mean_bp_dist = 0.5*np.mean([np.abs(x-y) for (x, y) in list(zip(i, j))])
    return mean_bp_dist

def is_PK(s):
    '''return if dotbracket structure represents a PK'''
    return ("[" in s) or ("{" in s) or ("<" in s)


def compare_structure_to_native(s, native, metric="all", PK_involved=None):
    if metric not in ["PPV", "sensitivity", "F1_score", "all"]:
        raise ValueError(
            'Only PPV, sensitivity, F1_score and all comparison currently implemented.')

    if PK_involved is None:
        s_list = convert_dotbracket_to_bp_list(s, allow_pseudoknots=True)
        native_list = convert_dotbracket_to_bp_list(
            native, allow_pseudoknots=True)
    elif PK_involved:
        s_list = _seperate_structure_into_PK_involved_or_not(s)["pk_bps"]
        native_list = _seperate_structure_into_PK_involved_or_not(native)[
            "pk_bps"]
    else:
        s_list = _seperate_structure_into_PK_involved_or_not(s)["no_pk_bps"]
        native_list = _seperate_structure_into_PK_involved_or_not(native)[
            "no_pk_bps"]

    PP = len(s_list)
    P = len(native_list)

    TP = len([x for x in s_list if x in native_list])  # true positives
    FP = len([x for x in s_list if x not in native_list])  # false positives
    FN = len([x for x in native_list if x not in s_list])  # false negative

    PPV = TP / PP if PP != 0 else 0
    sen = TP / P if P != 0 else 0
    F1 = (2 * PPV * sen) / (sen + PPV) if sen + PPV != 0 else 0

    if metric == "PPV":
        return PPV
    elif metric == "sensitivity":
        return sen
    elif metric == "F1_score":
        return F1
    elif metric == "all":
        return {"PPV": PPV, "sensitivity": sen, "F1_score": F1}


def compare_structures_to_natives(structs, natives, comparison="basepairs", metric="all"):
    # TODO other metrics? https://en.wikipedia.org/wiki/Positive_and_negative_predictive_values#Negative_predictive_value_(NPV)
    if comparison not in ["basepairs", "is_PK", "non_PK_basepairs", "PK_basepairs"]:
        raise ValueError(
            'Only basepairs and is_PK comparison currently implemented.')
    if metric not in ["PPV", "sensitivity", "F1_score", "all"]:
        raise ValueError(
            'Only PPV, sensitivity, F1_score and all comparison currently implemented.')

    if comparison == "basepairs":
        structs_list = [convert_dotbracket_to_bp_list(
            s, allow_pseudoknots=True) for s in structs]
        natives_list = [convert_dotbracket_to_bp_list(
            native, allow_pseudoknots=True) for native in natives]

        PP = sum(len(x) for x in structs_list)
        P = sum(len(x) for x in natives_list)

        TP = sum(len([z for z in x if z in y])
                 for x, y in zip(structs_list, natives_list))  # true positives

    elif comparison == "non_PK_basepairs":
        structs_list = [_seperate_structure_into_PK_involved_or_not(
            s)["no_pk_bps"] for s in structs]
        natives_list = [_seperate_structure_into_PK_involved_or_not(
            native)["no_pk_bps"] for native in natives]
        PP = sum(len(x) for x in structs_list)
        P = sum(len(x) for x in natives_list)

        TP = sum(len([z for z in x if z in y])
                 for x, y in zip(structs_list, natives_list))  # true positives

    elif comparison == "PK_basepairs":
        structs_list = [_seperate_structure_into_PK_involved_or_not(s)[
            "pk_bps"] for s in structs]
        natives_list = [_seperate_structure_into_PK_involved_or_not(
            native)["pk_bps"] for native in natives]
        PP = sum(len(x) for x in structs_list)
        P = sum(len(x) for x in natives_list)

        TP = sum(len([z for z in x if z in y])
                 for x, y in zip(structs_list, natives_list))  # true positives

    elif comparison == "is_PK":
        s_is_PK = [is_PK(s) for s in structs]
        native_is_PK = [is_PK(native) for native in natives]

        PP = sum(s_is_PK)
        P = sum(native_is_PK)

        # true positives
        TP = sum([x and y for x, y in zip(s_is_PK, native_is_PK)])
        # false positives
        FP = sum([x and not y for x, y in zip(s_is_PK, native_is_PK)])
        # false negatives
        FN = sum([not x and y for x, y in zip(s_is_PK, native_is_PK)])
        TN = sum([not x and not y for x, y in zip(
            s_is_PK, native_is_PK)])  # true negatives

    PPV = TP / PP if PP != 0 else 0
    sen = TP / P if P != 0 else 0
    F1 = (2 * PPV * sen) / (sen + PPV) if sen + PPV != 0 else 0

    if metric == "PPV":
        return PPV
    elif metric == "sensitivity":
        return sen
    elif metric == "F1_score":
        return F1
    elif metric == "all":
        return {"PPV": PPV, "sensitivity": sen, "F1_score": F1}


###############################################################################
# Structure helpers
###############################################################################


def _seperate_structure_into_PK_involved_or_not(s):
    bp_list = convert_dotbracket_to_bp_list(s, allow_pseudoknots=True)
    groups = _group_into_non_conflicting_bp(bp_list)
    bp_list_no_pk = groups[0]
    bp_list_pk = [bp for group in groups[1:] for bp in group]
    return {"no_pk_bps": bp_list_no_pk, "pk_bps": bp_list_pk}


def _get_non_redudant_bp_list(conflict_list):
    ''' given a conflict list get the list of nonredundant basepairs this list has

    Args:
            conflict_list: list of pairs of base_pairs that are intertwined basepairs
    returns:
            list of basepairs in conflict list without repeats
    '''
    non_redudant_bp_list = []
    for conflict in conflict_list:
        if conflict[0] not in non_redudant_bp_list:
            non_redudant_bp_list.append(conflict[0])
        if conflict[1] not in non_redudant_bp_list:
            non_redudant_bp_list.append(conflict[1])
    return non_redudant_bp_list


def _get_list_bp_conflicts(bp_list):
    '''given a bp_list gives the list of conflicts bp-s which indicate PK structure
    Args:
            bp_list: of list of base pairs where the base pairs are list of indeces of the bp in increasing order (bp[0]<bp[1])
    returns:
            List of conflicting basepairs, where conflicting is pairs of base pairs that are intertwined.
    '''
    if len(bp_list) <= 1:
        return []
    else:
        current_bp = bp_list[0]
        conflicts = []
        for bp in bp_list[1:]:
            if (bp[0] < current_bp[1] and current_bp[1] < bp[1]):
                conflicts.append([current_bp, bp])
        return conflicts + _get_list_bp_conflicts(bp_list[1:])


def _group_into_non_conflicting_bp(bp_list):
    ''' given a bp_list, group basepairs into groups that do not conflict

    Args
            bp_list: list of base_pairs

    Returns:
            groups of baspairs that are not intertwined
    '''
    conflict_list = _get_list_bp_conflicts(bp_list)

    non_redudant_bp_list = _get_non_redudant_bp_list(conflict_list)
    bp_with_no_conflict = [
        bp for bp in bp_list if bp not in non_redudant_bp_list]
    groups = [bp_with_no_conflict]
    while non_redudant_bp_list != []:
        current_bp = non_redudant_bp_list[0]
        current_bp_conflicts = []
        for conflict in conflict_list:
            if current_bp == conflict[0]:
                current_bp_conflicts.append(conflict[1])
            elif current_bp == conflict[1]:
                current_bp_conflicts.append(conflict[0])
        max_group = [
            bp for bp in non_redudant_bp_list if bp not in current_bp_conflicts]
        to_remove = []
        for i, bpA in enumerate(max_group):
            for bpB in max_group[i:]:
                if bpA not in to_remove and bpB not in to_remove:
                    if [bpA, bpB] in conflict_list or [bpB, bpA] in conflict_list:
                        to_remove.append(bpB)
        group = [bp for bp in max_group if bp not in to_remove]
        groups.append(group)
        non_redudant_bp_list = current_bp_conflicts
        conflict_list = [conflict for conflict in conflict_list if conflict[0]
                         not in group and conflict[1] not in group]
    return groups


###############################################################################
# ORPHANED unused anywhere in package
# figure out utility, take outside of utils and document or depricate
###############################################################################


def convert_multiple_dbns_to_eternafold_input(seq, list_of_constraint_strings, filename):
    '''hard-coded to have 3 constraints right now for use in eternafold training with kd-ligand data.'''
    constraint_list = []
    for constraint_string in list_of_constraint_strings:
        constraint_list.append(write_constraint_string(seq, constraint_string))

    with open('%s' % filename, 'w') as out:
        for i in range(len(seq)):
            out.write('%d\t%s\t%d\t%d\t%d\n' % (
                i+1, seq[i], constraint_list[0][i], constraint_list[1][i], constraint_list[2][i]))


def write_reactivity_file_vienna(reactivities, sequence, fname=None):
    '''write reactivity (either SHAPE or DMS) to file format used by ViennaRNA.

      ex:
      1 U 0.120768
      2 G 0.190510
      3 U 0.155776

    Args:
      reactivities (list): a list of normalized reactivity float data. 
      sequence: RNA sequence
      Negative numbers can be used to indicate no signal.
    '''

    assert len(reactivities) == len(sequence)

    if fname is None:
        fname = '%s.SHAPE' % filename()

    with open(fname, 'w') as f:
        i = 1
        for char, reactivity in list(zip(sequence, reactivities)):
            if reactivity >= 0:
                f.write('%d %s %f\n' % (i, char, reactivity))

            i += 1
    return fname


def get_missing_motif_bases(seq):
    FMN_apt1 = 'AGGAUAU'
    FMN_apt2 = 'AGAAGG'
    a = seq.find(FMN_apt1) - 1
    # print(a)
    b = seq.find(FMN_apt2) + len(FMN_apt2)
    #print(seq.find(FMN_apt2), b)

    return seq[a], seq[b]


def combo_list_to_dbn_list(seq, final_combo_list, apt_idx_list, apt_ss_list):
    """ Helper function for write_constraints 
    Converts the combination of aptamer fragments to dbn strings
    """
    dbn_string_list = []
    for combo in final_combo_list:
        dbn_string = ['.']*len(seq)
        temp = apt_idx_list[combo, :][:, 1:3]

        # Fill in the dbn string with the aptamer ss
        for (start, finish), apt_ss in zip(temp, apt_ss_list):
            dbn_string[start:finish] = list(apt_ss)
        dbn_string = ''.join(dbn_string)

        # Check if aptamer is flipped
        if temp[0, 0] > temp[-1, 0]:
            dbn_string = flip_ss(dbn_string)
        dbn_string_list.append(dbn_string)
    return dbn_string_list


def write_combo_constraints(seq, raw_apt_seq, raw_apt_ss, verbose=False):
    """ Given a sequence, get all possible secondary constraints of the aptamer 

    Args:
      seq: RNA sequence
      raw_apt_seq: aptamer sequence
        e.g. CAAAG+CAAAG+GGCCUUUUGGCC
        + denotes splitable aptamer
      raw_apt_ss: aptamer secondary structure
        e.g. (xxx(+)xxx)+((((xxxx))))
        + denotes splitable aptamer
        x denotes unpaired base
        . denotes wildcard (can be anything)
      verbose: to be verbose
    Returns
      list of all possible dbn_string for the given aptamer
    """
    if verbose:
        print('Writing constraints')
        print(seq)
        print(raw_apt_seq)
        print(raw_apt_ss)
    # split everything by +
    apt_seq_list = raw_apt_seq.split('+')
    apt_ss_list = raw_apt_ss.split('+')
    if len(apt_seq_list) != len(apt_ss_list):
        raise ValueError(
            'Missing + in aptamer sequence or secondary structure')

    # Iterate through each aptamer fragment and save its idx,
    apt_idx_list = []
    for idx, (apt_seq, apt_ss) in enumerate(zip(apt_seq_list, apt_ss_list)):
        if seq.find(apt_seq) == -1:
            raise ValueError("Aptamer segment {} not found".format(idx+1))
        if len(apt_seq) != len(apt_ss):
            raise ValueError(
                "Mismatch between aptamer sequence and aptamer secondary structure")

        # Save all locations of each fragment
        for m in re.finditer(apt_seq, seq):
            # Note: cannot get overlapping segments
            span = m.span()
            start = span[0]
            finish = span[1]
            ss_length = len(apt_seq)
            apt_idx_list.append([idx, start, finish, ss_length])
    apt_idx_list = np.array(apt_idx_list)

    # Now combine aptamer fragments into full secondary structure constraints
    N_frag = len(apt_ss_list)  # Number of fragments to stitch together

    # Get a list of all possible combination (each combination is a list as well)
    temp = np.array([np.where(apt_idx_list[:, 0] == idx)[0]
                     for idx in range(N_frag)])
    if len(temp) > 1:
        combo = np.meshgrid(*temp)
        # reformat the combinations
        combo_list = np.array(combo).T.reshape(-1, N_frag)
    else:
        combo_list = np.array([[x] for x in temp[0]])

    # Check each combination to make sure it's feasible, if not remove
    final_combo_list = prune_combo_list(combo_list, apt_idx_list, N_frag)

    # Convert each combination into a dbn_string
    dbn_list = combo_list_to_dbn_list(
        seq, final_combo_list, apt_idx_list, apt_ss_list)

    # List will be empty if nothing can be in order (only for >= 3 fragments)
    return dbn_list


def write_constraint_string(seq, constraint_dbn):
    '''write set of integers to represent constraints, i.e. for use in bpseq format.'''

    assert(len(seq) == len(constraint_dbn))

    bp_list = convert_dotbracket_to_bp_dict(constraint_dbn)

    constraint_list = []

    for i, c in enumerate(constraint_dbn):
        if c == 'x':
            constraint = 0
        elif c == '.':
            constraint = -1  # or -1 if undefined
        elif c in ['(', ')']:
            constraint = bp_list[i]+1
        else:
            print('Error reading constraint string', c)
        constraint_list.append(constraint)
    return constraint_list


def prune_combo_list(combo_list, apt_idx_list, N_frag):
    """ Helper function for write_constraints 
    Prunes all possible combinations of the aptamer fragments
    """
    final_idx_list = []
    for idx, combo in enumerate(combo_list):
        temp = apt_idx_list[combo, :]
        start_idx_list = temp[:, 1]
        ss_len_list = temp[:, 3]
        if N_frag == 1:  # If <= 2 fragments than always in order
            final_idx_list.append(idx)
        elif N_frag == 2:
            dx = np.diff(start_idx_list)
            if dx != 0:
                final_idx_list.append(idx)
        else:
            # first check if aptamer is in the correct order in the sequence
            dx = np.diff(start_idx_list)
            if np.all(dx < 0) or np.all(dx > 0):
                # second check if aptamer fragments don't overlap
                # each element in dx must be >= length of ss
                if dx[0] > 0:  # fragments in increasing order
                    ss_len_list = ss_len_list[:-1]
                else:  # fragments in decreasing order
                    ss_len_list = ss_len_list[1:]

                # Check if there is enough bases to fit the aptamer fragments
                if np.all(np.abs(dx) >= ss_len_list):
                    final_idx_list.append(idx)
    final_combo_list = combo_list[final_idx_list]
    return np.array(final_combo_list)


def flip_ss(ss):
    """ Flips a secondary structure 
    Only flips the unpaired bases
    e.g. (.((....))....( ->
         ).((....))....)
    """
    bp_list = []
    unmatch_list = []
    ss = list(ss)
    for idx, c in enumerate(ss):
        if c == '(':
            bp_list.append(idx)
        elif c == ')':
            if len(bp_list):
                bp_list.pop()
            else:
                unmatch_list.append(idx)
    unmatch_list += bp_list
    for idx in unmatch_list:
        if ss[idx] == '(':
            ss[idx] = ')'
        else:
            ss[idx] = '('
    ss = ''.join(ss)
    return ss

def write_constraints(seq, motif=False, MS2=False, LIG=False, lig1=('nAGGAUAU', '(xxxxxx('), lig2=('AGAAGGn', ')xxxxx)')):
    '''Inputs:
    seq: RNA sequence
    motif: tuple (seq, struct) of motif. For example: PUM would be motif=('UGUAUAUA','xxxxxxxx').
    MS2: bool, whether to include MS2 constraint or not
    lig1: tuple (seq, struct) for 5' portion of ligand aptamer. Default is FMN.
    lig2: tuple (seq, struct) for 3' portion of ligand aptamer

    Outputs:
    dbn string, () for paired, x for unpaired, . for unspecified
    '''

    # when FMN aptamer and MS2 aptamer overlap, MS2 loses out on bp
    MS2_apt = 'ACAUGAGGAUCACCCAUGU'
    LIG_apt1 = lig1[0].replace('n', '')
    LIG_apt2 = lig2[0].replace('n', '')

    unpaired_list = []
    bp_list = {}

    dbn_string = ['.']*len(seq)

    if motif:
        if LIG:
            raise ValueError(
                'Sorry, due to some hacky hard-coding, cannot handle both motif and LIG inputs at this time.')
        else:
            return write_constraints(seq, LIG=True, lig1=motif, lig2=('', ''))

    if LIG:
        if seq.find(LIG_apt1) == -1:
            raise RuntimeError("ligand 5' aptamer domain not found, %s" % seq)

        else:
            if lig1[0].startswith('n'):
                start1 = seq.find(LIG_apt1) + len(LIG_apt1) - len(lig1[0])
                if start1 < 0:  # hws: changed from 1, maybe wrong?
                    start1 = seq.find(
                        LIG_apt1, start1+len(lig1[0])+1) + len(LIG_apt1) - len(lig1[0])
            else:
                start1 = seq.find(LIG_apt1)
                if start1 < 0:  # hws: changed from 1, maybe wrong?
                    start1 = seq.find(LIG_apt1, start1+len(lig1[0])+1)

            finish1 = start1 + len(lig1[0])

            if lig2[0].startswith('n'):
                start2 = seq.find(LIG_apt2, finish1+1) + \
                    len(LIG_apt2) - len(lig2[0])
            else:
                start2 = seq.find(LIG_apt2, finish1+1)
            finish2 = start2 + len(lig2[0])
            #print('start1, finish1, start2, finish2 FMN', start1, finish1, start2, finish2)
            dbn_string[start1:finish1] = list(lig1[1])
            dbn_string[start2:finish2] = list(lig2[1])

    if MS2:
        if seq.find(MS2_apt) == -1:
            raise RuntimeError("MS2 aptamer domain not found: %s" % seq)
        else:
            start = seq.find(MS2_apt)
            finish = start+len(MS2_apt)
            #print('start, finish MS2', start, finish)

            if dbn_string[start] != ".":
                #print('warning, aptamer overlap')
                dbn_string[start+1:finish-1] = list('((((x((xxxx))))))')
            else:
                dbn_string[start:finish] = list('(((((x((xxxx)))))))')

    return ''.join(dbn_string)


import random
import numpy as np
from .mfe import mfe
from .utils import get_bpp_from_dbn
from .utils import filename
from .utils import load_package_locations
from os import remove

# load package locations from yaml file, watch! global dict
package_locs = load_package_locations()

def get_bootstrap_reac_file(reactivity):
    reac_file = '%s.SHAPE' % filename()
    range_arr = np.arange(1, len(reactivity) + 1)
    reac_arr = np.array(reactivity)
    shape_pos = np.array([range_arr, reac_arr]).T
    sample_idx = np.random.choice(len(reactivity), len(reactivity))
    shape_pos = shape_pos[sample_idx,:]

    f = open(reac_file, 'w')
    for cur_sample in shape_pos:
        pos, reactivity = cur_sample
        if reactivity > 0:
            f.write('%d %f\n' % (pos, reactivity))
    f.close()

    return reac_file

def mfe_bootstrap(seq, num_bootstrap, 
    package='rnastructure', T=37,
    constraint=None, shape_signal=None, dms_signal=None, pseudo=False):
    """
    Compute MFE structure (within package) for RNA sequence with bootstrapping on the SHAPE/DMS data.

        Args:
        seq (str): nucleic acid sequence
        T (float): temperature (Celsius)
        constraint (str): structure constraints
        shape_signal(list): list of normalized SHAPE reactivities, with negative values indicating no signal
        dms_signal(list): list of normalized DMS reactivities, with negative values indicating no signal
        pseudo: if True, will predict pseudoknots, but only with RNAstructure

        Possible packages: 
        'rnastructure'
        
    Returns
        string: MFE structure
        np array: Base-pair probability matrix from bootstrapping
    """
    if (shape_signal is None) and (dms_signal is None):
        raise ValueError("Bootstrapping only applies if you have reactivity data.")
    if package != 'rnastructure':
        raise ValueError("Bootstrapping only runs for now with RNAstructure")

    bpp_matrix = np.zeros((len(seq), len(seq)))

    mfe_struct = mfe(seq, package=package, T=T, constraint=constraint, 
        shape_signal=shape_signal, dms_signal=dms_signal, pseudo=pseudo)

    for bootstrap in range(num_bootstrap):
        shape_file = None
        dms_file = None
        
        if shape_signal is not None:
            shape_file = get_bootstrap_reac_file(shape_signal)
        if dms_signal is not None:
            dms_file = get_bootstrap_reac_file(dms_signal)

        cur_mfe_struct = mfe(seq, package=package, T=T, constraint=constraint, 
            shape_file=shape_file, dms_file=dms_file, pseudo=pseudo)
        bpp_matrix += get_bpp_from_dbn(cur_mfe_struct)

        if shape_signal is not None:
            remove(shape_file)
        if dms_signal is not None:
            remove(dms_file)

    return [mfe_struct, bpp_matrix/num_bootstrap]
